/* ========================================================================== *
 * Lib         : HtmlComb.c                                                   *
 * Description : Implementation file for the HtmlComb library.                *
 * Version     : 1.0.0 05/Mar/2004                                            *
 * ========================================================================== */

# include "Prelude.h"
# include "HtmlComb.h"
# include "html.c"
# include "html_cata.c"

/* == Combinators =========================================================== */

Html html_text (char *s)
{ return ConsHtml(ConsHtmlText(s),NoHtml()); }
                                                                                
Html html_tag (char *s , Html h)
{ return ConsHtml(ConsHtmlTag(s,NoHtmlAttrs(),h,s),NoHtml()); }
                                                                                
Html html_tagWithAttrs (char *s, HtmlAttrs attrs, Html h)
{ return ConsHtml(ConsHtmlTag(s,attrs,h,s),NoHtml()); }
                                                                                
HtmlAttr html_attr (char *n , char *v)
{ return ConsHtmlAttr (n,v); }
                                                                                
Html html_font (Html h) 
{ return html_tag ("font",h); }
                                                                                
Html html_bold (Html h) 
{ return html_tag ("b",h); }

Html html_italic (Html h) 
{ return html_tag ("i",h); }

Html html_emph (Html h) 
{ return html_tag ("em",h); }
                                                                                
Html html_hline (void) 
{ return html_tag ("hr", NoHtml()); }
                                                                                
Html html_par (Html h) 
{ return html_tag ("p", h); };
                                                                                
Html html_br (void) 
{ return ConsHtml(ConsHtmlText(" "), html_tag("br", NoHtml())); }
                                                                                
Html html_space (void) 
{ return html_text(" "); }
                                                                                
Html html_center (Html h) 
{ return html_tag ("center",h); }
                                                                                
Html html_table (Html h)
{ return html_tag ("table",h); }
                                                                                
Html html_row (Html h)
{ return html_tag ("tr",h); }
                                                                                
Html html_rowWithAttrs (HtmlAttrs attrs,Html h)
{ return html_tagWithAttrs("tr",attrs,h); }

Html html_elem (Html h)
{ return html_tag ("td",h); }
                                                                                
Html html_elemWithAttrs (HtmlAttrs attrs,Html h)
{ return html_tagWithAttrs("td",attrs,h); }
                                                                                
HtmlAttr html_aleft (void)
{ return html_attr("align","left"); }
                                                                                
HtmlAttr html_aright (void)
{ return html_attr("align","right"); }
                                                                                
HtmlAttr html_acenter (void)
{ return html_attr("align","center"); }
                                                                             
Html html_link (char *url,Html h)
{ return html_tagWithAttrs("a",
                           ConsHtmlAttrs(html_attr("href",url),
                           NoHtmlAttrs()),h); 
}
                                                                                
Html html_image (char *url,Html h)
{ return html_tagWithAttrs("img",
                           ConsHtmlAttrs(html_attr("src",url),
                           NoHtmlAttrs()),h); 
}

Html html_itemize (Html h)
{ return html_tag ("ul",h);  } ;
                                                                                
Html html_item (Html h)
{ return html_tag ("li",h);  } ;

/* == Unparse =============================================================== */

string unparse_HtmlLst(HtmlLst e)
{ string s;
  if (e->tag == sel_NoHtmlLst)
   return string_null() ;
  s = unparse_Html(e->u.ConsHtmlLst.p0);  
  s = string_cat(s,unparse_HtmlLst(e->u.ConsHtmlLst.p1));
  return s;
}


string unparse_Html(Html e)
{ string s;
  s = (string) cata_Html(e, either_Html(_u_f0_,_u_f1_));  
  s = string_cat(s, string_fromCStr("\n"));
  return s;
}

T _u_f0_(void)
{ return (T) string_null(); }

T _u_f1_(HtmlEl e, T t)
{ string a;
  string b;
  a = (string) cata_HtmlEl(e, either_HtmlEl(_u_f2_,_u_f3_));
  //a = string_cat(a, string_fromCStr("\n"));
  b = (string) t;
  return (T) string_cat(a,b); 
}

T _u_f2_(char *s)
{ return (T) string_fromCStr(s); }

T _u_f3_(char *s, HtmlAttrs la, Html h, char *n)
{ string a;
  a = string_fromCStr("<");
  a = string_cat(a, string_fromCStr(s));
  a = string_cat(a, (string)cata_HtmlAttrs(la,either_HtmlAttrs(_u_f4_,_u_f5_)));
  a = string_cat(a, string_fromCStr(">"));
  a = string_cat(a, unparse_Html(h));
  a = string_cat(a, string_fromCStr("</"));
  a = string_cat(a, string_fromCStr(n));
  a = string_cat(a, string_fromCStr(">"));
  return (T) a;
}

T _u_f4_(void)
{ return (T) string_null(); }

T _u_f5_(HtmlAttr h, T t)
{ string a;
  string b;
  a = string_fromCStr(" ");
  a = string_cat(a, cata_HtmlAttr(h,either_HtmlAttr(_u_f6_)));
  b = (string) t;
  return (T) string_cat(a,b); 
}

T _u_f6_(char *s, char *n)
{ string a;
  a = string_fromCStr(s);
  a = string_cat(a, string_fromCStr(" = \""));
  a = string_cat(a, string_fromCStr(n));
  a = string_cat(a, string_fromCStr("\""));   
  return (T) a; 
}

/* ========================================================================== */
