/* ========================================================================== *
 * Lib         : libPrelude.c                                                 *
 * Description : Implementation file for the Prelude library.                 *
 * Version     : 1.0.0 16/Jan/2004                                            *
 * ========================================================================== */

#include "Prelude.h"

/* == Primitive Types ======================================================= */

extern void * int_add(void * a,void * b) 
{ return ((void *)( ((int) a) + ((int) b) )); }

extern void * add_a_1(void * a,void * b)
{ return (int_succ(a)); }

extern void * add_1_a(void * a,void * b)
{ return (int_succ(b)); }

extern void * int_mul(void * a,void * b)
{ return ((void *)( ((int) a) * ((int) b) )); }

extern void * int_div(void * a,void * b)
{ return ((void *)( ((int) a) / ((int) b) )); }

extern void * int_sub(void * a,void * b)
{ return ((void *)( ((int) a) - ((int) b) )); }

extern Ordering int_compare(void *a,void *b)
{ if (((int) a) > ((int) b)) return GT;
   else if (((int) a) == ((int) b)) return EQL;
     else return LT; }

extern void * int_succ (void *a)
{ return ((void *)( ((int) a) + 1)); }

extern void * int_pred (void *a)
{ return ((void *)( ((int) a) - 1)); }

extern void * id(void * i)
{ return i; }

extern string bool_show (int b)
{ if (b) return string_fromCStr("True") ;
  else return string_fromCStr("False") ; 
}

extern string real_show (double b)
{ return string_fromCStr(make_string("%f",b)) ; }

extern string int_show (int b)
{ return string_fromCStr(make_string("%i",b)) ; }

extern string char_show (char b)
{ return string_fromCStr(make_string("%c",b)) ; }

/* == string ================================================================ */

extern string string_null (void)
{   string ast;

    ast = (string) malloc(sizeof (struct s_string));
    if (!ast) return (0);
    ast->tag = sel_string_null;
    return (ast);
};

extern string string_cons (char c,string next)
{   string ast;
    
    ast = (string) malloc(sizeof (struct s_string));
    if (!ast) return (0);
    ast->tag = sel_string_cons;
    ast->u.string_cons.c = c;
    ast->u.string_cons.next = next;
    return (ast);
};

extern efsstring either_string (T (*string_null) (void),
                                T (*string_cons) (char, T))
{ efsstring ast;
  
  ast = (efsstring) malloc(sizeof (struct s_efsstring));
  if (!ast) return (0);
  ast->string_null = string_null;
  ast->string_cons = string_cons;
  return (ast);
};

extern T string_cata (string a, efsstring f)
{ T r;

  switch (a->tag)
  { case (sel_string_null) : r = f->string_null();
                             break;
    case (sel_string_cons) : r = f->string_cons
                                 (a->u.string_cons.c,
                                  string_cata(a->u.string_cons.next,f));
                             break;
  };
  return (r);
};

extern string string_show (string a)
{ void * _show_null_string_ (void) ;
  void * _show_cons_string_ (char c, void *) ;
  return (string) string_cata (a,either_string(_show_null_string_,
                                               _show_cons_string_)) ; 
}

extern string string_clone (string a)
{ extern void * _clone_null_string_(void) ;
  extern void * _clone_cons_string_(char c, void *t) ;
  return (string) string_cata (a,either_string(_clone_null_string_,
                                               _clone_cons_string_)) ;
}

extern int string_length (string a)
{ extern void * _length_nil_string_(void) ;
  extern void * _length_cons_string_(char, void *) ;
  return (int) string_cata (a,either_string(_length_nil_string_,
                                            _length_cons_string_)) ;
}

extern string string_cat(string s,string a)
{  string aux,prev,ret;

   if (s->tag == sel_string_null)
     return (string_clone(a));

   if (a->tag == sel_string_null)
     return (string_clone(s));
   
   aux = string_clone(s);
   ret = aux;
   
   while (aux->tag != sel_string_null)
   { prev = aux;
     aux = aux->u.string_cons.next; }

   prev->u.string_cons.next = string_clone(a);
   return (ret);
}

extern int string_eq (string a, string b)
{ return (!(strcmp(string_toCStr(a),string_toCStr(b)))); }

extern char * string_toCStr (string a)
{ char * str ;
  char * start_str ;
  string pointer ;  

  pointer   = a ;
  start_str = str = (char *) malloc(1 + string_length (a)); 
  
  while (pointer->tag != sel_string_null)
  { *str = pointer->u.string_cons.c; 
    *str++;
    pointer = pointer->u.string_cons.next; }
  *str = '\0';
  return (start_str);  
}

extern string string_fromCStr(char * s)
{  string str = string_null() ;
   int    i   = strlen(s) - 1 ;

   while (i >= 0)
   { str = string_cons(s[i],str);
     i--; }
   return (str);
}

void * _show_null_string_ ()
{ return (void *) string_fromCStr ("string_null()") ; }
void * _show_cons_string_ (char c, void *t)
{ string a;
  a = string_null();
  a = string_cat(a, string_fromCStr("string_cons("));
  a = string_cat(a, string_cons(c,string_cons(',',string_null()))) ;
  a = string_cat(a, (string) t);
  a = string_cat(a, string_fromCStr(")"));
  return (void *) a;
}
void * _length_nil_string_ ()
{ return (void *) 0 ; }
void * _length_cons_string_ (char c, void *t)
{ return (void *) 1 + ((int) t) ; }
void * _clone_null_string_()
{ return ((void *)string_null()) ;}
void * _clone_cons_string_(char c, void *t)
{ return ((void *)string_cons(c,(string)t)); }

/* == usefull functions ===================================================== */

char * make_string(const char *fmt, ...) 
{
   int n, size = 100;
   char *p;
   va_list ap;
   if ((p = malloc (size)) == NULL)
      return NULL;
   while (1) 
   {
     va_start(ap, fmt);
     n = vsnprintf (p, size, fmt, ap);
     va_end(ap);
     if (n > -1 && n < size)
      return p;

     if (n > -1)    /* glibc 2.1 */
      size = n+1; 
     else           /* glibc 2.0 */
      size *= 2;  
     if ((p = realloc (p, size)) == NULL)
      return NULL;
   }
};

/* ========================================================================== */
