#include <stdio.h>
#include "lrclib.h"
#include <EVALUATOR/evaluator.h>

#ifdef VISUAL
#include "lrcedit.h"
#endif

#ifdef VISUAL
/*debug info output flags*/
int set_print_events(int);
int set_time_events(int); 
int tk_program_set_verbose(int);
int vsd_set_verbose(int);
#endif

static int EMIT_ETF=0;
static int LOAD_ETF=0;
static int show_tree=0;
static int print_indented=0;
static int show_output_packet=0;
static int trace_visits=0;



/*** 
 ***  READING INFORMATION ABOUT AG
 ***
 ***/
static char * search_path[] = { ".", "/users/matthys/d.lrc/SSL_C", 0 };
static int have_ag_tables=0;

#ifdef USER_ACTIONS
extern lrc_bool b_trace_equations;
#endif

#ifdef USER_ACTIONS
void init_ag_tables() {
  if (have_ag_tables) return;
  if (!lrc_read_ag_tables(search_path)) {
     printf("Could not read data of attribute grammar\n");
     exit(1);
  }
  allocate_eq_trace_data(ag_num_equations());
  have_ag_tables=1;
}
#endif

extern void lrc_set_ip_attr(lrc_term *ip, char * name, lrc_term attr_value);
typedef struct output_descriptor {
  struct output_descriptor *next;
  lrc_string attr;
  lrc_string view;
  lrc_string file;
} *OutputDescriptor;

static OutputDescriptor all_outputs=0;

static void enter_output_descriptor(lrc_string attr, lrc_string view, lrc_string file) {
  OutputDescriptor od;

  od=(OutputDescriptor)lrc_malloc(sizeof(*od));

  printf("entering output descriptor: attr='%s' view='%s' file='%s'\n", attr, view, file);
  od->attr=attr; od->view=view; od->file=file;
  od->next=all_outputs;
  all_outputs=od; 
}

#ifdef TOKEN_CONVERSION
#include "token_conversion.c"
#endif

void LrcMatchFailure() {
  printf("Matching failed!!!\n");
  printf("  this is probably fatal...\n\n");
}

int lrc_printf(lrc_string s) {
  printf("%s",s);
  return 1;
}

float ticks_to_seconds(int num_ticks) {
  float ticks_per_second=1000000.0;
  return num_ticks/ticks_per_second;
}

static void unparse_decorated_term(DecoratedTerm dt, lrc_string view_name, lrc_string file_name) {
  UnparseDescriptor ud;
  Reference ref;
  ref=create_root_reference(dt->ast);
  
  if (dt==0) return;
  ud=dt_sr_view2unparse_descriptor(dt,ref,view_name);
  unparse_to_file(ud,file_name);
}

static void unparse_dt_path(DecoratedTerm dt, lrc_string view_name, Path p, lrc_string file_name) {
  UnparseDescriptor ud;
  Reference ref;
  ref=create_root_reference(dt->ast);
  ref->sub0=p;
  
  if (dt==0) return;
  ud=dt_sr_view2unparse_descriptor(dt,ref,view_name);
  unparse_to_file(ud,file_name);
}

/** Unparse a term to a FILE
 **/
static void emit_unattributed_term(lrc_term t, lrc_string view_name, FILE *f) {
   UnparseDescriptor ud;

   ud.term=t;
   ud.uat=0;
   ud.view_name=view_name;
   unparse_to_file_stream(ud,f);
}

static lrc_term get_root_op_attr(char * attr_name, lrc_term *op)
{ int   attr_index;

  attr_index=lrc_attr_get_op_index(lrc_root_nonterminal(), 1 /* vis_num */, attr_name);
  if (attr_index==-1) { 
    printf("INTERNAL ERROR: get_root_op_attr: could not locate attribute %s\n", attr_name );
    return 0;
  }
  return op[attr_index];
}


/** Unparse a named root attribute to a file 
 **/
static void emit_root_attribute(DecoratedTerm dt, lrc_string attr_name, lrc_string view, FILE *f) {
  lrc_term attr;
  lrc_term * op;

  op  = dt_get_decoration(dt);
  if (op==0) return;

  attr=get_root_op_attr(attr_name,op);
  if (attr!=0) {
    emit_unattributed_term(attr,view,f);
  }
}

static int emit_according_to_output_descriptor(DecoratedTerm dt, OutputDescriptor od) {
  FILE *f;
  if (od->file==0) {
     f=stdout; 
  }
  else {
     f=fopen(od->file,"w");
     if (!f) {
       printf("Cannot open %s\n",od->file);
       return 0;
     }
  }
  if (od->attr!=0) {
    emit_root_attribute(dt,od->attr,od->view,f);
  }
  else {
    printf("Unparse decorated term\n");
    unparse_decorated_term(dt,od->view,od->file);
  }
}


lrc_term compute_abstract_tree(lrc_term concrete_tree) {
  lrc_term * op;
  op=lrc_visit_term_1(concrete_tree);
  if (op==0) return 0;
  return op[0];
}

extern lrc_term lrc_parse_file(char * fname, int nt);
extern lrc_term lrc_yacc_parse_string(char * s, int nt);

void explain_parse_error(void) {
   printf("Parser failed: ");
   switch (lrc_parse_state()) {
      case PS_OK: break;
      case PS_FILE_NOT_READABLE: printf("file is not readable"); break;
      case PS_NO_ENTRY: printf("nonterminal has no parse entry"); break;
      case PS_SYNTAX_ERROR: printf("syntax error"); break; 
   }
   printf("\n");
}

lrc_term string_to_concrete(char * s, int nt) {
  lrc_term parse_tree=lrc_yacc_parse_string(s,nt);
  if (parse_tree==0) explain_parse_error();
  return parse_tree;
}

lrc_term file_to_concrete(char * fname,int nt) {
  lrc_term parse_tree=lrc_parse_file(fname,nt);
  if (parse_tree==0) explain_parse_error();
  if (show_tree) {
    printf("parse_tree=");
    if (print_indented) { printf("\n"); lrc_term_print_indent(parse_tree,1); }
    else { lrc_term_print_long(parse_tree); }
    printf("\n"); 
  }
  return parse_tree;
}

lrc_term concrete_to_abstract(lrc_term concrete_tree) {
  lrc_term * op;
  if (concrete_tree==0) return 0;

  /* printf("Computing abstract syntax tree\n"); */
  lrc_memo_start_AE();
  op=lrc_visit_term_1(concrete_tree);
  lrc_memo_end_AE();
  if (op==0) return 0;
  return op[0];
}

DecoratedTerm decorate_ast(lrc_term abs_tree) {
  DecoratedTerm res;
  
  if (abs_tree==0) return 0;

#ifdef TRACE
  if (trace_visits) set_packet_printing();
#endif

  /** printf("IN decorate ast:\n"); **/
  res=create_decorated_term();
  /** printf("  dt_enter_ast\n"); **/
  dt_enter_ast(res,abs_tree);
  /** printf("  dt_decorate_term\n"); **/
  dt_decorate_term(res);
  
#ifdef TRACE
  if (trace_visits) unset_packet_printing();
#endif

  /** printf("end of decorate ast\n"); **/
  return res;
}

void process_decorated_term(DecoratedTerm dt) {
   lrc_term ast;
   lrc_term * op;
   int op_size, ix;

   if (dt==0) return;
   
   ast = dt_edse_initial_ast(dt);
   op  = dt_get_decoration(dt);
   if (op==0 || ast==0) return;
   if (all_outputs!=0) {
      OutputDescriptor od;
      od=all_outputs;
      while (od!=0) {
        emit_according_to_output_descriptor(dt,od);
        od=od->next;
      }
   }
   if (!show_output_packet) return;
   op_size=lrc_prod_vnum_op_size(LRC_TAG(ast),1);
   for(ix=0;ix<op_size;ix++) {
     lrc_string attr_name;
     attr_name=lrc_attr_get_op_name(lrc_root_nonterminal(),1,ix);
     printf("output_packet[%d]=(%s)",ix,attr_name);
     lrc_term_print_long(op[ix]); printf("\n");
   }
}

#if VISUAL==2
static void editor_install_term(lrc_term ast) {
  editor_enter_new_term(ast);
}
#endif

void process_abstract_tree(lrc_term ast) {
  if (ast==0) return;
#ifdef VISUAL 
  editor_install_term(ast);
#else
  process_decorated_term(decorate_ast(ast));
#endif
}

static int emit_etf(char * etf_name, lrc_term t) {
  FILE * etf;
  etf=fopen(etf_name,"w");
  if (!etf) {
    printf("Cannot open %s\n",etf_name);
    return 0;
  }
  else {
    lrc_term_store_etf(etf,t);
    return 1;
  }
}

static lrc_string infile_to_etf_file(char * fname) {
  lrc_string res;
  res=lrc_vstring_concat(fname,".etf",0);
  return res;
}

static int lrc_process_file(char * fname, int nt) {
  lrc_term concrete;
  lrc_term abstract;

  if (LOAD_ETF) {
   etf_structure abstract_etf;
   abstract_etf = load_etf(fname);
   if (abstract_etf.status == 1) {
    abstract = abstract_etf.term;
   }
   else {
    printf("Error: %s\n", abstract_etf.message);
    printf("Line : %s\n", abstract_etf.etf_line);
    abstract = 0;   
   }
  }
  else {
   concrete=file_to_concrete(fname,nt);
   if (concrete==0) return 0;
   abstract=concrete_to_abstract(concrete);
  }


  if (show_tree) {
      printf("abstract tree=");
      if (abstract==0) printf("(null)\n");
      else {
        if (print_indented) {
           printf("\n");
           lrc_term_print_indent(abstract,1);
        }
        else {
           lrc_term_print_long(abstract); 
        }
        printf("\n");
      }
  }
  if (abstract==0) return 0;
  if (EMIT_ETF) {
      emit_etf(infile_to_etf_file(fname),abstract);
  }
  process_abstract_tree(abstract);
  return 1;
}

void lrc_process_string(char * str,int nt) {
  process_abstract_tree(concrete_to_abstract(string_to_concrete(str,nt)));
}

static void Init(void) {
  lrc_lib_init();
  lrc_memo_init();
  init_grammar_tables();
  init_unparser();
  /*** lrc_set_string_parser(lrc_yacc_parse_string); **/
#ifdef TRANSFORM
  /* printf(" defining transformations...\n"); */
  lrc_define_transformations();
#endif
#ifdef VISUAL
# if VISUAL==2
  term_editor_init();
# else
  editor_init();
#endif
#endif
}

void explain_options(void) {
  printf ("Usage: eva [option] file\n");
  printf ("Valid options are:\n");
#ifdef VISUAL
  printf ("  -C display generated Tcl/Tk code\n");
  printf ("  -X1 trace visualizer\n");
  printf ("  -X2 deep trace of visualizer\n");
  printf ("  -E display event information\n");
  printf ("  -t display timing info\n");
#endif
  printf ("  -T show trees that are processed\n");
  printf ("  -i use indentation when printing trees\n");
  printf ("  -f load etf file\n");
  printf ("  -F emit etf file\n");
  printf ("  -P show computed output packets\n");
  printf ("  -Y trace parsing\n");
  printf ("  -V trace visits\n");
#ifndef VISUAL
  printf ("  -O write tree or attribute of root node to file\n");
  printf ("       -O<attr_name>\n");
  printf ("       -O<attr_name>:<view_name>\n");
  printf ("       -O<attr_name>::<file_name>\n");
  printf ("       -O<attr_name>:<view_name>:<file_name>\n");
  printf ("       -O:<view_name>\n");
  printf ("       -O:<view_name>:<file_name>\n");
#endif
  printf ("  -M define memoization strategy\n");
  printf ("       -M<strategy>,<number of collection-free evaluation>,<number of caches>\n");
  printf ("  -S define memoization tree-size\n");
  printf ("       -S<use tree size>,<minimal tree size>\n");
  printf ("  -h print this message\n");
}

void process_option_memo(char *str)
{ int a,b,c;
  sscanf(str,"-M%d,%d,%d",&a,&b,&c);
  /*** lrc_memo_strategy_config(a,b,c);  ***/
}


void process_option_tree_size(char *str)
{ int a,b;
  sscanf(str,"-S%d,%d",&a,&b);
  /*** lrc_memo_tree_size_config(a,b); ***/
}

static p_lrc_string_list split_string(lrc_string s, char c) {
  p_lrc_string_list res;
  Textbuffer tb;
  
  if (s==0) { return 0; }; 
  tb=textbuffer_init();
  while (*s!=0 && *s!=c) {
    textbuffer_append_char(tb,*s);
    s++;
  }
  if (*s==0) {
    return lrc_strings_2(textbuffer_retrieve(tb),0);
  }
  else {
     /* *s==c */
    return lrc_strings_2(textbuffer_retrieve(tb),split_string(s+1,c));
  }
}

/** s has the form -O<attr_name>
      or           -O<attr_name>:<view_name>
      or           -O<attr_name>::<file_name>
      or           -O<attr_name>:<view_name>:<file_name>
 **/
static void handle_output_option(lrc_string s) {
  p_lrc_string_list strs;
  p_lrc_string_list strs_tmp;
  lrc_string attr, view, file;

  s=s+2; /* skip the leading '-O' */
  strs=split_string(s,':');

 /*****
  {  p_lrc_string_list strs_tmp=strs;
     printf("strings are");
     while (strs_tmp!=0) {
       printf(" '%s'", strs_tmp->str);
       strs_tmp=strs_tmp->next;
     }
     printf("\n");
  }
 *****/
 
  if (strs==0) {
    printf("Could not handle option '%s'\n",s);
    return;
  }
  attr=strs->str;
  strs=strs->next;
  if (strs==0 || strs->str==0) {
     view="BASEVIEW";
  }
  else {
    view=strs->str;
  }
  if (strs!=0) {
    strs=strs->next;
  }
  file=0;
  if (strs!=0) {
    file=strs->str;
  }
  enter_output_descriptor(attr,view,file);
}

/**  s has form -x or -x<num>
     return      0 or the value of <num>
***/
static int get_option_number(char * s) {
  s=s+2;
  if (s==0) return 0;
  return STRtoINT(s);
}

void process_option(char *str) {
  switch(str[1]) {
#      ifdef VISUAL
	case 'X' : vsd_set_verbose(get_option_number(str)); break;
	case 'C' : tk_program_set_verbose(1);break;
	case 'E' : set_print_events(1);break;
        case 't' : set_time_events(1); break;
#     endif
#     ifdef TOKEN_CONVERSION
        case 'l' : define_token_convertor(token_convertor); break;
#     endif
        case 'T' : show_tree=1; break;
        case 'i' : print_indented=1; break;
        case 'F' : EMIT_ETF=1; break;
        case 'f' : LOAD_ETF=1; break;
        case 'P' : show_output_packet=1; break;
	case 'Y' : set_parse_debug(1); break;
        case 'V' : trace_visits=1; set_packet_printing(); break;
        case 'h' : explain_options(); exit(1); break;
	case 'M' : process_option_memo(str);
		   break;
	case 'S' : process_option_tree_size(str);
		   break;
        case 'O' : handle_output_option(str);
                   break;
        default : printf ("Unknown option.\n");
                  explain_options();
		  exit(1);
  } 
}

int
#ifdef DEBUGCODE
  evaluator_main
#else
  main
#endif
(
  int argc,
  char **argv
)
{
  int arg_num;

  Init();
  for (arg_num=1;arg_num<argc;arg_num++) {
    if (argv[arg_num][0]=='-') {  /* an option */
      process_option(argv[arg_num]);
    }
   else {

     printf("/**** processing file %s ****/\n", argv[arg_num]);
     lrc_process_file(argv[arg_num],lrc_root_nonterminal());

   }
  }
# ifdef VISUAL
  editor_accept_user_actions();
# endif
}


