#include <stdio.h>
#include <string.h>
#include <lrclib.h>
#include <EVALUATOR/evaluator.h>

int yydebug ;

#include "bcode.ttm"
#include "bcode.htc"


static int trace_scanner=0;
static int convert_tokens=0;

static tp_f_token_convertor the_token_convertor;

void define_token_convertor(tp_f_token_convertor convertor) {
  convert_tokens=1;
  the_token_convertor=convertor;
}


int yylex(void);
int yyparse(void);
void yyerror(char * mess) {
  printf("YYERROR: %s\n", mess);
}

static int first_call;
static int lrc_entry_symbol;

static lrc_term lrc_parse_tree;

void init_parser(int entry) {
  first_call=1;
  lrc_entry_symbol=entry;
  lrc_clear_parse_position();
}

extern int yyleng;
extern char * yytext;

int lrc_yylex(void) {
  int res;
  if (first_call) {
    first_call=0;
    return lrc_entry_symbol;
  }
  res = yylex();
  if (trace_scanner) {
      /* printf("yyleng=%d\n", yyleng); */
      printf("SCANNED: <%s> (code=%d; token length=%d)\n", yytext, res, yyleng);
  }
  if (convert_tokens) {
      res=the_token_convertor(res,yytext);
  }
  return res;
}

#define yylex() lrc_yylex()

#define lrc_no_white() ("")
#include "bcode.tab.c"

void set_parse_debug(int n) {
  yydebug=n;
  trace_scanner=n;
}

static lrc_term lrc_parse(int entry_token) {
  int parse_result;

  /* printf("Starting parser...\n");
   */
  init_parser(entry_token);
  parse_result=yyparse();
  if (parse_result==0) { /* OK! */
    /* printf("End of parser...\n");
     * printf("Root tag is %d\n",LRC_TAG(lrc_parse_tree));
     * printf("Number of sons is %d\n",lrc_constructor_num_sons(LRC_TAG(lrc_parse_tree)));
     * printf("Address of root is %d\n",(int)lrc_parse_tree);
     */
   return lrc_parse_tree;
  }
  else {
    int line, col;
    lrc_set_parse_state(PS_SYNTAX_ERROR);
    lrc_get_parse_position(&line,&col);
    printf("Parse error at position (%d,%d)\n", line, col);
    return 0;
  }
}

void lrc_set_scanner_start_state(char * ss);
void lrc_pop_scanner_start_state(void);

static void prepare_scanner_for_nonterminal(int nt) {
  char * scanner_state;
  scanner_state=lrc_nont_scan_state(nt);
  if (!( lrc_string_equal(scanner_state,"") ||
         lrc_string_equal(scanner_state,"LRC_NO_SCAN_STATE")
       )
     ) {
    /* printf("Setting scanner state to %s\n",scanner_state); */
    lrc_set_scanner_start_state(scanner_state);
  }
}

static void close_scanner_for_nonterminal(int nt) {
  char * scanner_state;
  scanner_state=lrc_nont_scan_state(nt);
  if (!( lrc_string_equal(scanner_state,"") ||
         lrc_string_equal(scanner_state,"LRC_NO_SCAN_STATE")
       )
     ) {
    /* printf("Popping scanner state\n"); */
    lrc_pop_scanner_start_state();
  }
}

extern FILE * yyin;
static int num_processed_files=0;

lrc_term lrc_parse_file(char * fname, int nt) {
  lrc_term p;
  int entry;

  lrc_set_parse_state(PS_OK);
  yyin=fopen(fname,"r");
  if (yyin==0) {
    lrc_set_parse_state(PS_FILE_NOT_READABLE);
    return 0;
  }
  if (num_processed_files>=1) yyrestart(yyin);
  num_processed_files++;
  entry=lrc_nont_entry_token(nt);
  if (entry==-1) { /* No entry */
    lrc_set_parse_state(PS_NO_ENTRY);
    return 0;
  }
  prepare_scanner_for_nonterminal(nt);
  p=lrc_parse(entry);
  fclose(yyin);
  close_scanner_for_nonterminal(nt);
  if (p!=0) return LRC_CHILD(p,1);
  return 0;
}


#ifndef yywrap
int yywrap() { return 1; }
#endif

/* Parsing from strings */
typedef void * YY_BUFFER_STATE;
extern YY_BUFFER_STATE yy_scan_string(char *);
extern void yy_delete_buffer(YY_BUFFER_STATE);

lrc_term lrc_yacc_parse_string(char * s, int nt) {
  YY_BUFFER_STATE buffer_handle;
  lrc_term p;
  int entry; /* entry token */

  lrc_set_parse_state(PS_OK);
  /* printf("Parsing. Nonterminal is %s(%d)\n", lrc_nont_ident(nt), nt); */
  entry=lrc_nont_entry_token(nt);
  if (entry==-1) { /* No entry */
    lrc_set_parse_state(PS_NO_ENTRY);
    return 0;
  }

  prepare_scanner_for_nonterminal(nt);
  buffer_handle=yy_scan_string(s);
  p=lrc_parse(entry);
  if (p!=0) p=LRC_CHILD(p,1);
  close_scanner_for_nonterminal(nt);

  yy_delete_buffer(buffer_handle);
  /* printf("End of Parsing.\n"); */
  return p;
}
