### Default bindings
bind Text <Control-f> {zoek_eerste %W }
bind Text <Control-n> {zoek_volgende %W}


### Creation of widgets

### Button
proc create_button { wname wnumber caption } {
  ##puts "!!!!!!>>> create_button '$wname' '$number' '$caption'"

  set_widget_number $wname $wnumber
  button $wname -text $caption
  $wname configure -command [list handle_push_button $wname]
  configure_button $wname $wnumber
}

proc configure_button { wname wnumber } {
  ##puts "configure_button '$wname' '$wnumber'"
  foreach pname {width height background foreground activebackground activeforeground relief state bitmap} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

proc handle_push_button { wname } {
  set wnumber [get_widget_number $wname]
  ##puts "handle_push_button $wname"
  if ![catch {widget_get_property $wnumber ACTION} action] {
     ##puts "Yes! Widget '$wname' has action '$action'"
     $action $wname
  } else {
    ##puts "	NO! Widget has no associated ACTION: $action"
    filter_events "B" $wnumber
  }

}

### Text
proc create_text { wname wnumber } {
  set_widget_number $wname $wnumber
  text $wname -wrap none -exportselection 0 -borderwidth 2 -relief sunken
  $wname tag configure structsel -underline 1 -background gray
  configure_text $wname $wnumber
}

proc configure_text { wname wnumber } {
  ##puts ">>configure_text '$wname' '$wnumber'"
  foreach pname {width height background foreground insertbackground} {
    if ![catch {widget_get_property $wnumber "$pname"} pvalue] {
     ##puts "configuring '$wname': $pname to $pvalue"
     $wname configure -$pname $pvalue
    }
  }
  ##puts "Setting red foreground"
  ##$wname configure -foreground red
}

### Menubutton
proc create_menubutton { wname wnumber caption} {
  set_widget_number $wname $wnumber
  ##puts "create_menu_button '$wname' '$wnumber' '$caption'"
  menubutton $wname -text $caption
  configure_menubutton $wname $wnumber
}

proc configure_menubutton { wname wnumber } {
  ##puts ">>>configure_menubutton '$wname' '$wnumber'"
  foreach pname {background activebackground foreground relief bitmap} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}
### Message
proc create_message { wname wnumber mess} {
  set_widget_number $wname $wnumber
  ##puts "create_menu_button '$wname' '$wnumber' '$caption'"
  menubutton $wname -text $mess
  configure_message $wname $wnumber
}

proc configure_message { wname wnumber } {
  ##puts ">>>configure_menubutton '$wname' '$wnumber'"
  foreach pname {background activebackground foreground text width} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

### Entry
proc create_entry { wname wnumber } {
  set_widget_number $wname $wnumber
  entry $wname
  widget_set_property_default $wname exportselection 0
  widget_set_property_default $wname relief groove
  widget_set_property_default $wname width 10
  ##$wname tag configure structsel -underline 1 -background gray
  configure_entry $wname $wnumber
}

proc configure_entry { wname wnumber } {
  ##puts "configure_menubutton '$wname' '$wnumber'"
  foreach pname {background activebackground foreground exportselection relief} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

#### Scale
proc create_scale { wname wnumber from_value to_value } {
  global var$wname
  set_widget_number $wname $wnumber
  ##puts "create_scale '$wname' '$wnumber' '$caption'"
  scale $wname -from $from_value -to $to_value -variable var$wname
  widget_set_property_default $wname orient horizontal
  bind $wname <ButtonRelease-1> [list handle_scale_event $wname $wnumber]
  configure_scale $wname $wnumber
  ## The following code is necessary to let tk first set the screen presentation of the scale;
  ## before calling filter_events.
  bindtags $wname [list Scale $wname . all ]
}

proc handle_scale_event { wname wnumber } {
  filter_events "l" $wnumber [$wname get]
}

proc configure_scale { wname wnumber } {
  ##puts "configure_scale '$wname' '$wnumber'"
  global var$wname
  foreach pname {label orient} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
  if ![catch {widget_get_property $wnumber value} pvalue] {
     ##puts "setting variable 'var$wname' of scale to $pvalue"
     set var$wname $pvalue
     ##$wname configure -$pname $pvalue
  }
}

#### Label
proc create_label { wname wnumber label_text } {
  set_widget_number $wname $wnumber
  label $wname -text $label_text
  configure_label $wname $wnumber
}

proc configure_label { wname wnumber } {
  ##puts "configure_label '$wname' '$wnumber'"
  foreach pname {font} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

#### Label
proc create_listbox { wname wnumber } {
  set_widget_number $wname $wnumber
  listbox $wname -relief sunken -width 10 -height 10 -selectmode single
  configure_listbox $wname $wnumber
}

proc configure_listbox { wname wnumber } {
  ##puts "configure_listbox '$wname' '$wnumber'"
  foreach pname {font relief orient width height selectmode} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

### Pulldown menu
proc configure_pulldownmenu { wname wnumber } {
  ##puts "configure_pulldown '$wname' '$wnumber'"
  ##puts "configure_scale '$wname' '$wnumber'"
  foreach pname {width height background} {
    if ![catch {widget_get_property $wnumber $pname} pvalue] {
     $wname configure -$pname $pvalue
    }
  }
}

### Menu entry
proc configure_menu_entry { encl_wname entry_wnumber entry_position } {
  ##puts "configure_menu_entry '$encl_wname' '$entry_wnumber' '$entry_position'"
  foreach pname {background activebackground foreground activeforeground accelerator underline} {
    if ![catch {widget_get_property $entry_wnumber $pname} pvalue] {
     ##puts "  setting '$pname' to '$pvalue'"
     $encl_wname entryconfigure $entry_position -$pname $pvalue
    }
  }
}

############################################################
proc set_widget_number { wname wnumber } {
  global widget_info
  set widget_info($wname,number) $wnumber
}

proc get_widget_number { wname } {
  global widget_info
  return $widget_info($wname,number)
}


### Text editing
bind Text <Key>		   { text_typed_char %W %A }
bind Text <Delete>	   { text_delete_forward_char %W }
bind Text <Control-Key-Return>	{ text_parse_text %W }
bind Text <ButtonPress-1> { text_mouse_b1_click %W [%W index @%x,%y] }
bind Text <ButtonPress-2> { text_mouse_b2_click %W [%W index @%x,%y] }
bind Text <ButtonPress-3> { text_mouse_b3_click %W %X %Y }
bind Text <Control-Key-z> { text_undo %W }
bind Text <Control-Key-p> { structural_to_parent %W }
bind Text <Control-Key-f> { structural_forward %W }
bind Text <Control-Key-b> { structural_backward %W }
bind Text <Control-Key-x> { text_close_editor %W }
bind Text <Control-Key-m> { lrc_report_memory_usage }
bind Text <Control-Key-u> { lrc_print_unparse_fragments [get_widget_number %W] }
##bind Text <Control-Key-o> { text_control %W }
bind Text <Control-Key-o> { open_file_dialog %W }
bind Text <Control-Key-s> { save_to_file %W}
bind Text <Control-Key-t> { save_term_to_file %W}
bind Text <Control-Key-f> { save_widget_to_file %W}
bind Text <Control-Key>	  { text_control %W }
bind Text <BackSpace>	  { text_delete_backward_char %W }
bind Text <Control-Key-h> { text_delete_backward_char %W }
bind Text <Control-Key-d> { text_duplicate_editor %W }
bind Text <Tab>		  { start_command %W }
bind Text <Key-Down>	  { widget_cursor_down %W }
bind Text <Key-Up>	  { widget_cursor_up %W }
bind Text <Key-Left>	  { widget_cursor_left %W }
bind Text <Key-Right>	  { widget_cursor_right %W }
bind Text <Key-Home>	  { widget_cursor_line_start %W }
bind Text <Key-End>	  { widget_cursor_line_end %W }



bind Entry <Key>	    { entry_typed_char %W %A }
bind Entry <Delete>	    { entry_delete_forward_char %W }
bind Entry <Control-Key-Return>	 { entry_parse_text %W }
bind Entry <ButtonPress-1> { entry_mouse_b1_click %W [%W index @%x] }
bind Entry <ButtonPress-2> { entry_mouse_b2_click %W [%W index @%x] }
bind Entry <ButtonPress-3> { entry_mouse_b3_click %W %X %Y }
bind Entry <Control-Key-z> { entry_undo %W }
bind Entry <Control-Key-p> { structural_to_parent %W }
bind Entry <Control-Key-f> { structural_forward %W }
bind Entry <Control-Key-b> { structural_backward %W }
bind Entry <Control-Key-x> { entry_close_editor %W }
bind Entry <Control-Key-m> { lrc_report_memory_usage }
bind Entry <Control-Key-u> { lrc_print_unparse_fragments [get_widget_number %W] }
##bind Entry <Control-Key-o> { entry_control %W }
bind Entry <Control-Key-o> { open_file_dialog %W }
bind Entry <Control-Key-s> { save_to_file %W}
bind Entry <Control-Key-t> { save_term_to_file %W}
bind Entry <Control-Key>   { entry_control %W }
bind Entry <BackSpace>	   { entry_delete_backward_char %W }
bind Entry <Control-Key-h> { entry_delete_backward_char %W }
bind Entry <Control-Key-d> { entry_duplicate_editor %W }
bind Entry <Tab>	   { start_command %W }
bind Entry <Key-Left>	   { entry_cursor_left %W }
bind Entry <Key-Right>	   { entry_cursor_right %W }
bind Entry <Key-Home>	   { widget_cursor_line_start %W }
bind Entry <Key-End>	   { widget_cursor_line_end %W }

# bookkeeping for textual editing
proc widget_ute { W } {
  global aute
  if [info exists aute($W,us_start)] {
    return 1
  } else {
    return 0
  }
}

proc open_text_editing { W range_start range_end } {
  global aute
  if [info exists aute($W,us_start)] { return }
  set aute($W,us_start) $range_start
  set aute($W,us_end)	$range_end
  set aute($W,ute) 1
  set aute($W,original_text) [$W get $range_start $range_end]
}

proc open_entry_editing { W range_start range_end } {
  global aute
  if [info exists aute($W,us_start)] { return }
  ##puts "open entry editing: $range_start; $range_end"
  set aute($W,us_start) $range_start
  set aute($W,us_end)	$range_end
  set aute($W,ute) 1
  set aute($W,original_text) [$W get]
}

proc close_text_editing { W } {
  global aute
  ##puts "Closing text editing"
  unset aute($W,us_start)
  unset aute($W,us_end)
  unset aute($W,original_text)
  unset aute($W,ute)
  foreach {start end} [$W tag ranges buffer] {
    $W tag remove buffer $start $end
  }
}

proc close_entry_editing { W } {
  global aute
  ##puts "Closing entry editing"
  unset aute($W,us_start)
  unset aute($W,us_end)
  unset aute($W,original_text)
  unset aute($W,ute)

  ##configure_entry $wname [get_widget_number $wname]
}

# start textual editing of structural selection in text widget W
proc text_start_editing { W } {
  global aute
  if [info exists aute] {
    if [info exists aute($W,ute)] return;
  }
  set ss_range [$W tag nextrange structsel 1.0]
  if {[llength $ss_range] == 0} {
    set range_start [$W index insert]
    set range_end $range_start
  } else {
    set range_start [lindex $ss_range 0]
    set range_end   [lindex $ss_range 1]
  }
  puts "text_start_editing $range_start $range_end"
  ## todo: verify whether text editing is allowed...
  ## supposing it is allowed:
  $W tag configure buffer -underline 0 -background yellow -foreground red;
  $W tag add buffer $range_start $range_end;
  $W tag raise buffer
  open_text_editing $W $range_start $range_end
  return 1
}

# start textual editing of structural selection in entry widget W
proc entry_start_editing { W } {
  global aute
  if [info exists aute] {
    if [info exists aute($W,ute)] return;
  }
  ##set ss_range [$W tag nextrange structsel 1.0]
  set ss_range [list 0 [$W index end]]
  if {[llength $ss_range] == 0} {
    set range_start [$W index insert]
    set range_end $range_start
  } else {
    set range_start [lindex $ss_range 0]
    set range_end   [lindex $ss_range 1]
  }
  ##puts "entry_start_editing: lrc getting range in '$W'"
  set ss_range_new [structsel_range [get_widget_number $W]]
  ##puts "entry_start_editing: lrc says range in '$W' is $ss_range"
  ##puts "entry_start_editing $range_start $range_end"
  ## todo: verify whether text editing is allowed...
  ## supposing it is allowed:
  ##$W tag configure buffer -underline 0 -background yellow -foreground red;
  ##$W tag add buffer $range_start $range_end;
  ##$W tag raise buffer
  $W configure -background yellow -foreground red;
  open_entry_editing $W $range_start $range_end
  return 1
}

proc entry_tag_all_in_style { wname stylename } {

}

proc text_typed_char { W C } {
  set p [$W index insert]
  ##puts "User typed '$C' in '$W' at '$p'"
  if [expr [string compare $C ""] == 0] {
    ## a Control-nothing or Shift-nothing
    return
  }
  text_start_editing $W
  $W insert insert $C {buffer}
}

proc entry_typed_char { W C } {
  set p [$W index insert]
  ##puts "User typed '$C' in '$W' at '$p'"
  if [expr [string compare $C ""] == 0] {
    ## a Control-nothing or Shift-nothing
    return
  }
  entry_start_editing $W
  $W insert insert $C
}

proc text_delete_backward_char { W } {
  set p [$W index insert]
  ##puts "User-delete-backward-char in '$W' at '$p'"
  text_start_editing $W
  $W delete "insert -1c"
}

## delete character following cursor
proc text_delete_forward_char { W } {
  set p [$W index insert]
  ##puts "User-delete-forward-char in '$W' at '$p'"
  text_start_editing $W
  $W delete insert
}

proc entry_delete_forward_char { W } {
  set p [$W index insert]
  ##puts "User-delete-forward-char in '$W' at '$p'"
  entry_start_editing $W
  $W delete insert
}

proc text_control { W } {
  set p [$W index insert]
  ##puts "User-text-control in '$W'"
}

proc entry_control { W } {
  set p [$W index insert]
  ##puts "User-entry-control in '$W'"
}

proc text_parse_text { W } {
  if ![widget_ute $W] {
    return
  }
  global aute
  set p [$W index insert]
  puts "User-parse-text in '$W' at '$p'"
  set buffer_tagged [$W tag ranges buffer]
  if {[llength $buffer_tagged] == 0} {
    set new_text ""
  } else {
    set new_text [$W get buffer.first buffer.last]
  }
  if [replace_range_by_string [get_widget_number $W] $aute($W,us_start) $aute($W,us_end) $new_text] {
    close_text_editing $W
  } else {
    puts "oei: replacement failed..."
  }
}

proc entry_parse_text { W } {
  if ![widget_ute $W] {
    return
  }
  global aute
  set p [$W index insert]
  puts "User-parse-text in entry '$W' at '$p'"
  ##set buffer_tagged [$W tag ranges buffer]
  ##if {[llength $buffer_tagged] == 0} {
  ##  set new_text ""
  ##} else {
  ##  set new_text [$W get buffer.first buffer.last]
  ##}
  set new_text [$W get]
  if [replace_range_by_string [get_widget_number $W] $aute($W,us_start) $aute($W,us_end) $new_text] {
    close_entry_editing $W
  } else {
    puts "oei: replacement failed..."
  }
}

proc replace_range_by_string { W us_s us_e new_text } {
  puts "replace_range_by_string '$W' '$us_s' '$us_e' by "
  puts "   '$new_text'"
  if ![catch {widget_replace_range $W $us_s $us_e $new_text} mess] {
    advance_after_parse $W
    return 1
  } else {
    puts "oei: replacement error: $mess"
    return 0
  }
}

proc text_undo { W } {
  global aute
  if [widget_ute $W] {
    puts "editor undo:"
    puts $aute($W,original_text)
  }
}

proc text_control { W } {
  puts "STUB: text_control '$W'"
}

### Structural selection
proc text_set_structsel { W range_start range_end } {
  ##puts "text_set_structsel '$W' $range_start $range_end"
  $W tag remove structsel 1.0 end;
  $W tag add structsel $range_start $range_end;
  ##puts "text_set_structsel end"
}

proc structural_to_parent { W } {
  puts "structural_to_parent '$W'"
  ascend_to_parent [get_widget_number $W]
}

proc structural_forward { W } {
  puts "STUB: structural_forward '$W'"
}

proc structural_backward { W } {
  puts "STUB: structural_backward '$W'"
}

proc advance_after_parse { W } {
  puts "STUB: advance_after_parse '$W'"
}

################
proc text_mouse_b1_click { W pos } {
##  puts "mouse click-1 in '$W' at '$pos'"
  set_text_cursor $W $pos
  focus $W
}

proc text_mouse_b2_click { W pos } {
##  puts "STUB: mouse click-2 in '$W' at '$pos'"
}

proc text_mouse_b3_click { W X Y } {
##  puts "mouse click-3 in '$W' at '$X.$Y'"
  global trafo
  set trafo(X) $X
  set trafo(Y) $Y
  get_transformations [get_widget_number $W]
  ##puts "presenting transformations: number is $trafo(NUM)"
  if [expr $trafo(NUM) == 0] {
    return
  }
  tk_popup .lrc_trafos $trafo(X) $trafo(Y) 1
  set trafo(NUM) 0
}

###################################################
proc entry_mouse_b1_click { W pos } {
  ## puts "mouse click-1 in '$W' at '$pos'"
  set_entry_cursor $W $pos
  focus $W
}

proc entry_mouse_b2_click { W pos } {
  ##puts "STUB: mouse click-2 in '$W' at '$pos'"
}

proc entry_mouse_b3_click { W X Y } {
  ##puts "mouse click-3 in '$W' at '$X.$Y'"
  global trafo
  set trafo(X) $X
  set trafo(Y) $Y
  get_transformations [get_widget_number $W]
  ##puts "presenting transformations: number is $trafo(NUM)"
  if [expr $trafo(NUM) == 0] {
    return
  }
  tk_popup .lrc_trafos $trafo(X) $trafo(Y) 1
  set trafo(NUM) 0
}

## transformations
proc set_transformations { W names } {
  global trafo
  ##puts "Transformations in $W:"
  .lrc_trafos delete 0 last
  set trafo(NUM) [llength $names]
  foreach { trafo_name } $names {
    ##puts "  >>$trafo_name"
    .lrc_trafos add command -label $trafo_name -command [list  tk_perform_transformation $W $trafo_name ]
  }
}

proc tk_perform_transformation { W trafo_name } {
  ##puts "perform transformation $trafo_name from $W"
  perform_transformation [get_widget_number $W] $trafo_name
}


#### Diagnostic Information
proc create_diagnostic_display { } {
  global diagnostics
  if [info exists diagnostics(widget)] { return }
  set diagnostics(widget) 1
  label ._lrc_diagnostics -text "Usefull information"
  pack ._lrc_diagnostics
}

proc structural_info { nont prod } {
  create_diagnostic_display
  ._lrc_diagnostics configure -text "Selection is $nont : $prod"
}

#### Command processing
proc start_command { W } {
  ##puts "start_command '$W'"
}

#### File IO
proc edit_file { fname } {

}

proc open_file_dialog { W } {
  set file [fileDialog $W open]
  puts "File to open is '$file'"
  open_file $file normal
}

proc save_term_to_file { W } {
  set file [fileDialog $W save]
  if [string compare $file ""] {
    puts "save term to file '$file'"
    lrc_save_term_to_file [get_widget_number $W] $file
  }
}

proc save_to_file { W } {
  set file [fileDialog $W save]
  if [string compare $file ""] {
    puts "save to file '$file'"
    lrc_save_to_file [get_widget_number $W] $file
  }
}


proc save_widget_to_file { W } {
  set file [fileDialog $W save]
  if [string compare $file ""] {
    puts "save to file '$file'"
    lrc_unparse_widget_to_file [get_widget_number $W] $file
  }
}

proc save_to_file_dialog { W } {
  set file [fileDialog $W save]
  puts "Save to file '$file'"
}

proc fileDialog {w operation} {
    #	Type names	Extension(s)	Mac File Type(s)
    #
    #---------------------------------------------------------
    set types {
    {"Text files"	{.txt .doc} }
    {"Text files"	{}	TEXT}
    {"Tcl Scripts"	{.tcl}	    TEXT}
    {"Term files"	{.etf}	    }
    {"C Source Files"	{.c .h}	    }
    {"All Source Files" {.tcl .c .h}	}
    {"Image Files"	{.gif}	    }
    {"Image Files"	{.jpeg .jpg}	}
    {"Image Files"	""	{GIFF JPEG}}
    {"All files"	*}
    }
    if {$operation == "open"} {
    set file [tk_getOpenFile -filetypes $types -parent $w]
    } else {
    set file [tk_getSaveFile -filetypes $types -parent $w \
	-initialfile Untitled -defaultextension .txt]
    }
    return $file
}

###
proc widget_set_property_default { wname property_name default_value } {
  set wnumber [get_widget_number $wname]
  set pvalue [widget_get_property_default $wnumber $property_name $default_value]
  $wname configure -$property_name $pvalue
}


proc widget_get_property_default { wnumber property_name default_value } {
  if ![catch {widget_get_property $wnumber $property_name} pvalue] {
     return $pvalue
  } else {
    return $default_value
  }
}

###
proc text_duplicate_editor { W } {
  duplicate_editor [get_widget_number $W]
}


### Close and exit
proc text_close_editor { W } {
   close_editor [get_widget_number $W]
}

### Exit editor


####  Application specific proc's

proc do_advanced { W } {
  ##puts "Executing do_advanced from $W"
}

proc select_from_button { W } {
  ##puts "Select from button '$W'"
  widget_select_origin [get_widget_number $W]
}

proc menuentry_select { wnumber menu_label } {
  ##puts ">>!!>>menuentry_select '$wnumber' '$menu_label'";
  if ![catch {widget_get_property $wnumber ACTION} action] {
    if ![catch {$action $wnumber} action_result] {
       puts "Action succeeded"
    } else {
       puts "Failure in action: '$action_result'"
    }
  } else {
    puts "    NO! Menu entry '$menu_label' has no associated ACTION: $action"
    ##puts " filter_events l $wnumber $menu_label"
    filter_events "l" $wnumber $menu_label
  }
  ##puts ">>!!>>menuentry_select '$wnumber' '$menu_label' END OF SHOW"
}

     ##puts "Yes! Menu entry '$menu_label' has action '$action'"; $action
     ##puts "Performing action $action";
     ##$action $wnumber $menu_label
     ##exec $action
     #set action [widget_get_property $wnumber "ACTION"]
     ##exec_a_command [widget_get_property $wnumber "ACTION"]

proc exec_a_command { command } {
  print "Executing command '$command'"
  $command
  return
}

proc action_struct_sel_from_menu { wnumber  } {
  widget_selection_to_property $wnumber "menu_origin"
}

proc action_open_file { wnumber } {
  open_file_dialog "."
}


proc widget_cursor_down { wname } {
  set cur_pos [$wname index insert]
  set pos [$wname index "insert +1l"]
  set end_pos [$wname index end]
  if [$wname compare $pos < $cur_pos] {
    ##puts "   not moving right: new position before old"
    return
  }
  if [$wname compare $pos == $end_pos] {
    ##puts "   not moving right: new position is end-position"
    return
  }
  set_text_cursor $wname $pos
}

proc widget_cursor_up { wname } {
  set cur_pos [$wname index insert]
  set pos [$wname index "insert -1l"]
  set_text_cursor $wname $pos
}

proc widget_cursor_left { wname } {
  set pos [$wname index "insert -1c"]
  set_text_cursor $wname $pos
}

proc entry_cursor_left { wname } {
  set pos [$wname index insert]
  set new_pos [expr $pos - 1]
  set_entry_cursor $wname $new_pos
}

proc widget_cursor_right { wname } {
  set cur_pos [$wname index insert]
  set pos [$wname index "insert +1c"]
  set end_pos [$wname index end]
  if [$wname compare $pos < $cur_pos] {
    ##puts "   not moving right: new position before old"
    return
  }
  if [$wname compare $pos == $end_pos] {
    ##puts "   not moving right: new position is end-position"
    return
  }
  set_text_cursor $wname $pos
}

proc entry_cursor_right { wname } {
  set cur_pos [$wname index insert]
  set pos [expr $cur_pos + 1]
  set end_pos [$wname index end]
  if [expr $pos < $cur_pos] {
    ##puts "   not moving right: new position before old"
    return
  }
  if [expr $cur_pos == $end_pos] {
    ##puts "   not moving right: new position is end-position"
    return
  }
  set_entry_cursor $wname $pos
}

proc widget_cursor_line_start { wname } {
  set pos [$wname index "insert linestart"]
  set_text_cursor $wname $pos
}

proc widget_cursor_line_end { wname } {
  set pos [$wname index "insert lineend"]
  set_text_cursor $wname $pos
}

proc set_text_cursor { wname pos } {
  ##puts "set_text_cursor $wname TO $pos"
  if [widget_ute $wname] {
    ##puts "setting cursor during text editing to $pos"
    text_edit_set_cursor $wname $pos
  } else {
    set_cursor [get_widget_number $wname] $pos
  }
}

proc set_entry_cursor { wname pos } {
  ##puts "set_text_cursor $wname TO $pos"
  if [widget_ute $wname] {
    ##puts "setting cursor during text editing to $pos"
    entry_edit_set_cursor $wname $pos
  } else {
    set_cursor [get_widget_number $wname] $pos
  }
}

proc _tk_text_set_cursor { wname pos } {
  prefer_focus $wname
  $wname mark set insert $pos
}

proc _tk_entry_set_cursor { wname pos } {
  ##puts "setting cursor in entry '$wname' to '$pos'"
  prefer_focus $wname
  $wname icursor $pos
}

## setting the cursor during text editing

proc text_edit_set_cursor { wname pos } {
  set in_ute 0
  foreach {start end} [$wname tag ranges buffer] {
    ##puts "!! ute is $start--$end; pos is $pos"
    if [$wname compare $start <= $pos] {
      if [$wname compare $pos <= $end] {
	set in_ute 1
	break
      }
    }
  }
  if [expr $in_ute == 1] {
     ##puts "text_edit_set_cursor: position $pos inside ute"
     $wname mark set insert $pos
  }
}

proc entry_edit_set_cursor { wname pos } {
  set in_ute 0
  foreach {start end} [$wname tag ranges buffer] {
    ##puts "!! ute is $start--$end; pos is $pos"
    if [$wname compare $start <= $pos] {
      if [$wname compare $pos <= $end] {
	set in_ute 1
	break
      }
    }
  }
  if [expr $in_ute == 1] {
     ##puts "text_edit_set_cursor: position $pos inside ute"
     $wname mark set insert $pos
     _tk_entry_set_cursor $wname $pos
  }
}

## maintaining the focus
##    prefer_focus wname: set the focus to wname if no other window has the focus
##    force_focus wname: set the focus to wname
proc prefer_focus { wname } {
  global focus_info
  if [info exists focus_info(have_focus)] {
    return
  } else {
    force_focus $wname
  }
}

proc force_focus { wname } {
  global focus_info
  set focus_info(have_focus) $wname
  focus $wname
}


###########################################################################
########## User specific stuff (should be in a user-tcl-file)
proc open_an_file { wnumber } { open_file_dialog . }

proc application_structure_parent { wnumber } { ascend_to_parent $wnumber }

proc application_skip { wnumber } {	 }
proc application_exit { wnumber } { exit }

###########################################################################
########## Initialization

## this proc is the first one called from the c-editor
proc tk_editor_init { } {
  global trafo
  set trafo(NUM) 0
  wm title . "xxxx"
}

