#include <stdio.h>
#include <string.h>

#include "lrcmemory.h"
#include "ag_table.h"

static ag_rec_tp ag_table;

void allocate_attr_identifier_data(int num_idents) {
  /** printf("ALLOCATING attr_idents (%d)\n", num_idents); fflush(stdout); **/
  ag_table.num_attr_idents=num_idents;
  ag_table.ag_attr_idents
  =(identifier_tp *)lrc_malloc(num_idents*sizeof(*(ag_table.ag_attr_idents)));
}

void define_attr_ident(int ix, int id_code, char * s) {
  /** printf("Defining attr ident[%d], id_code=%d, ident=%s\n", ix, id_code, s); fflush(stdout); **/
  ag_table.ag_attr_idents[ix].ident=s;
  ag_table.ag_attr_idents[ix].id_code=id_code;
}

void print_attr_idents() {
  int i;
  for(i=0;i<ag_table.num_attr_idents;i++) {
    printf("Attr ident %d %s\n", i, ag_table.ag_attr_idents[i].ident ); fflush(stdout);
  }
}

void ag_print_attr_kind(attr_kind_tp akind) {
  if (akind==INH) { printf("INH"); return; }
  if (akind==SYN) { printf("SYN"); return; }
  if (akind==LOCAL) { printf("LOCAL"); return; }
}


char * index_ident_table(int ident_code) {
  int ix;
  for (ix=0; ix<ag_table.num_attr_idents; ix++) {
    if (ag_table.ag_attr_idents[ix].id_code==ident_code) {
       return ag_table.ag_attr_idents[ix].ident;
    }
  }
  printf("CANNOT LOCATE IDENTIFIER FOR CODE '%d'\n", ident_code); fflush(stdout);
  return "???";
}

void allocate_nonterminal_data(int num_nts) {
  ag_table.num_nonterminals=num_nts;
  ag_table.nonterminals=(nont_tp *)lrc_malloc(num_nts*sizeof(*(ag_table.nonterminals)));
}

void allocate_production_data(int num_prods) {
  /** printf("Allocating %d productions\n", num_prods); **/
  ag_table.num_productions=num_prods;
  ag_table.productions=(prod_tp *)lrc_malloc(num_prods*sizeof(*(ag_table.productions)));
}

void allocate_ntocc_data(int num_noccs) {
  ag_table.num_ntoccs=num_noccs;
  ag_table.ntoccs=(nocc_tp *)lrc_malloc(num_noccs*sizeof(*(ag_table.ntoccs)));
}

void allocate_attribute_data(int num_attrs) {
  ag_table.num_attributes=num_attrs;
  ag_table.attributes=(attr_tp *)lrc_malloc(num_attrs*sizeof(*(ag_table.attributes)));
}

void allocate_aocc_data(int num_aoccs) {
  ag_table.num_aoccs=num_aoccs;
  ag_table.aoccs=(aocc_tp *)lrc_malloc(num_aoccs*sizeof(*(ag_table.aoccs)));
}

void allocate_eq_data(int num_eqs) {
  /** printf("Allocating space for %d equations\n", num_eqs); **/
  ag_table.num_eqs=num_eqs;
  ag_table.eqs=(eq_tp *)lrc_malloc(num_eqs*sizeof(*(ag_table.eqs)));
}

void enter_nonterminal(int number, char * id) {
  /* printf("enter_nonterminal:%d|->%s\n", number,id); */
  ag_table.nonterminals[number].nont_id=id;
}

void enter_production(char * id, int number, int num_ntoccs, int first_ntocc) {
  /** printf("Enter prod %s at location '%d'\n", id, number); **/
  ag_table.productions[number].prod_id=id;
  ag_table.productions[number].num_ntoccs=num_ntoccs;
  ag_table.productions[number].first_ntocc=first_ntocc;
}

void enter_ntocc(int number, int prod, int nont, int multiplicity, int is_nta ) {
  ag_table.ntoccs[number].prod=prod;
  ag_table.ntoccs[number].nont=nont;
  ag_table.ntoccs[number].multiplicity=multiplicity;
  ag_table.ntoccs[number].is_nta=is_nta;
}

void enter_attribute(int ident_num, int number, int prod_nont, attr_kind_tp kind) {
  /***
    printf("Entering attribute ident_num='%d', number='%d', prod_nont='%d', kind='",
                             ident_num,      number,      prod_nont); fflush(stdout);
         ag_print_attr_kind(kind);
  printf("'\n"); fflush(stdout);
  ***/
  ag_table.attributes[number].ident_num=ident_num;
  ag_table.attributes[number].kind=kind;
  ag_table.attributes[number].prod_nont=prod_nont;
}

void enter_aocc_a(int number, int lntocc, int attr) {
  ag_table.aoccs[number].kind=AO_NORMAL;
  ag_table.aoccs[number].prod_ntocc=lntocc;
  ag_table.aoccs[number].attribute=attr;
  ag_table.aoccs[number].defining_eq=-1;
}

void enter_aocc_l(int number, int prod, int attr) {
  ag_table.aoccs[number].kind=AO_LOCAL;
  ag_table.aoccs[number].prod_ntocc=prod;
  ag_table.aoccs[number].attribute=attr;
  ag_table.aoccs[number].defining_eq=-1;
}

void enter_aocc_n(int number, int ntocc) {
  ag_table.aoccs[number].kind=AO_NTA_SR;
  ag_table.aoccs[number].prod_ntocc=ntocc;
  ag_table.aoccs[number].attribute=-1;
  ag_table.aoccs[number].defining_eq=-1;
}

void enter_eq(int number, int destination) {
  /** printf("Setting destination of equation '%d' to '%d'\n", number, destination); **/
  ag_table.eqs[number].destination=destination;
  ag_table.aoccs[destination].defining_eq=number;
}

#define NONT_FIELD(nont,fld) ag_table.nonterminals[nont].fld
#define NTOCC_FIELD(ntocc,fld) ag_table.ntoccs[ntocc].fld
#define PROD_FIELD(prod,fld) ag_table.productions[prod].fld
#define ATTR_FIELD(attr,fld) ag_table.attributes[attr].fld
#define AOCC_FIELD(aocc,fld) ag_table.aoccs[aocc].fld
#define EQ_FIELD(eq,fld) ag_table.eqs[eq].fld

int ag_num_productions()  { return ag_table.num_productions; }
int ag_num_nonterminals() { return ag_table.num_nonterminals; }
int ag_num_ntoccs()       { return ag_table.num_ntoccs; }
int ag_num_attributes()   { return ag_table.num_attributes; }
int ag_num_aoccs()        { return ag_table.num_aoccs; }
int ag_num_equations()    { return ag_table.num_eqs; }

char * nont_id(short nont) { return NONT_FIELD(nont,nont_id); }

short  ntocc_nont         (short ntocc) { return NTOCC_FIELD(ntocc,nont); }
short  ntocc_multiplicity (short ntocc) { return NTOCC_FIELD(ntocc,multiplicity); }
short  ntocc_prod         (short ntocc) { return NTOCC_FIELD(ntocc,prod); }
char * ntocc_id           (short ntocc) { return nont_id(ntocc_nont(ntocc)); }

char * prod_id(short prod) { return PROD_FIELD(prod,prod_id); }
short  prod_num_ntoccs(short prod) { return PROD_FIELD(prod,num_ntoccs); }
short  prod_first_ntocc(short prod) { return PROD_FIELD(prod,first_ntocc); }

short  prod_nth_ntocc(short prod, int n) { return PROD_FIELD(prod,first_ntocc)+n-1; }

char *       attr_id  (short attr) {
  /** printf("THIS IS attr_id(%d)\n", attr); fflush(stdout); **/
  return index_ident_table(ATTR_FIELD(attr,ident_num)); 
}
short        attr_nont(short attr) { return ATTR_FIELD(attr,prod_nont); }
short        attr_prod(short attr) { return ATTR_FIELD(attr,prod_nont); }
attr_kind_tp attr_kind(short attr) { return ATTR_FIELD(attr,kind); }

aocc_kind_tp aocc_kind(short aocc) { return AOCC_FIELD(aocc,kind); }
int   aocc_is_nta (short aocc) { return aocc_kind(aocc)==AO_NTA_SR; }
short aocc_attr   (short aocc) { return AOCC_FIELD(aocc,attribute); }
short aocc_nocc   (short aocc) { return AOCC_FIELD(aocc,prod_ntocc); }
short aocc_prod   (short aocc) { 
  if (aocc_kind(aocc)==AO_LOCAL) return AOCC_FIELD(aocc,prod_ntocc);
  return ntocc_prod(aocc_nocc(aocc));
}
short aocc_defining_eq(short aocc) {
  return AOCC_FIELD(aocc,defining_eq);
}

short eq_destination(short eq) {
  /** printf("eq_destination %d\n", eq); fflush(stdout); **/
  return EQ_FIELD(eq,destination);
}

void ag_print_prod(short prod) {
  short num_occs;
  short nocc_pos;

  num_occs=prod_num_ntoccs(prod);
  for(nocc_pos=1;nocc_pos<=num_occs; nocc_pos++) {
     printf("%s ", ntocc_id(prod_nth_ntocc(prod,nocc_pos)) );
     if (nocc_pos==1) printf(": %s( ", prod_id(prod));
  }
  printf(")\n");
}

void ag_print_aocc(short aocc) {
  short attr;
  short ao_kind;

  ao_kind=aocc_kind(aocc);
  if (ao_kind==AO_NTA_SR) { /* NTA */
     short nocc=aocc_nocc(aocc);
     printf("NTA-SR: %s:%s$%d", prod_id(aocc_prod(aocc)), ntocc_id(nocc), ntocc_multiplicity(nocc));
     return;
  }
  attr=aocc_attr(aocc);
  if (ao_kind==AO_NORMAL) {
    /*******************
    printf("ag_print_aocc NORMAL\n"); fflush(stdout);
    printf("AA\n"); fflush(stdout);
    printf("    attr=%d\n",attr); fflush(stdout);
    printf("BB\n"); fflush(stdout);
    printf("    attr_nont=%d\n",attr_nont(attr) ); fflush(stdout);
    printf("CC\n"); fflush(stdout);
    printf("    nont_id=%s\n",nont_id(attr_nont(attr))); fflush(stdout);
    printf("DD\n"); fflush(stdout);
    printf("    aocc_prod=%d\n",aocc_prod(aocc)); fflush(stdout);
    printf("EE\n"); fflush(stdout);
    printf("    aocc_nocc=%d\n",aocc_nocc(aocc)); fflush(stdout);
    printf("FF\n"); fflush(stdout);
    printf("    ntocc_multiplicity(aocc_nocc(aocc))=%d\n",ntocc_multiplicity(aocc_nocc(aocc))); fflush(stdout);
    printf("GG\n"); fflush(stdout);
    printf("    attr_id=%s\n",attr_id(attr)); fflush(stdout);
    printf("HH\n"); fflush(stdout);
    **************************/
    printf( "%s:%s$%d.%s",
             prod_id(aocc_prod(aocc)), nont_id(attr_nont(attr)),
             ntocc_multiplicity(aocc_nocc(aocc)), attr_id(attr) );
    return;
  }
  if (ao_kind==AO_LOCAL) {
    /*** printf("ag_print_aocc LOCAL\n"); fflush(stdout); ***/
    printf( "%s.%s", prod_id(attr_prod(attr)), attr_id(attr) );
    return;
  }
  printf( "Unknown kind for attribute %d:%d", attr, attr_kind(attr) );
}

void ag_print_equations() {
  int eq_num;
  for(eq_num=0;eq_num<ag_table.num_eqs;eq_num++) {
     printf("Equation '%d' defines ", eq_num); ag_print_aocc(eq_destination(eq_num));
     printf("\n");
  }
}

int ag_check_tables() {
  int status;

  status=1;
  return status;
}

