#include <stdio.h>
#include "lrcmemory.h"
#include "avl.h"

#define INLINE

#define HEIGHT(t) t->height
#define HASH(t) t->hash
#define LEFT(t) t->left
#define KEY(t) t->key
#define VALUE(t) t->value
#define RIGHT(t) t->right

static lrc_term lookup (lrc_term, AVL_tp, compare_function_tp);
static int      slope  (AVL_tp);
static AVL_tp   rebal_dot_create(AVL_tp, lrc_term, lrc_term, AVL_tp);
static AVL_tp   rotr   (AVL_tp);
static AVL_tp   rotl   (AVL_tp);
static AVL_tp   rotr_dot_create(AVL_tp, lrc_term, lrc_term, AVL_tp);
static AVL_tp   rotl_dot_create(AVL_tp, lrc_term, lrc_term, AVL_tp);
static AVL_tp   shiftr_dot_create(AVL_tp, lrc_term, lrc_term, AVL_tp);
static AVL_tp   shiftl_dot_create(AVL_tp, lrc_term, lrc_term, AVL_tp);


#define Max(a,b) ((a)>(b)?(a):(b))
#define fheight(t) ((t)==0?0:HEIGHT(t))

int avl_hash(AVL_tp avl) {
  return avl==0?0:avl->hash;
}

INLINE static int slope(AVL_tp avl){
  return  (avl==0?0:fheight(LEFT(avl))-fheight(RIGHT(avl)));
}

int avl_num_elements(AVL_tp avl){
  return avl==0?0: 1+ avl_num_elements(LEFT(avl))+avl_num_elements(RIGHT(avl));
}

void avl_forall(AVL_tp a, void action()) {
  if (a!=0) {
     avl_forall(LEFT(a),action);
     action(KEY(a),VALUE(a));
     avl_forall(RIGHT(a),action);
  }
}

void avl_forall_keys(AVL_tp a, void action()) {
  if (a!=0) {
     avl_forall(LEFT(a),action);
     action(KEY(a));
     avl_forall(RIGHT(a),action);
  }
}

/*
 *  Constructors for avl-trees;
 *
 */
AVL_tp avl_empty() { return 0; }

static AVL_tp create_bin(AVL_tp lft, lrc_term key, lrc_term val, AVL_tp rgt) {
  AVL_tp res;
  res         = (AVL_tp)ALLOC_AVL(sizeof(*res));
  HEIGHT(res) = 1+Max(fheight(lft),fheight(rgt));
  HASH(res)
  = (avl_hash(lft)+avl_hash(rgt)+lrc_term_hashix(key)+lrc_term_hashix(val)
    ) % lrc_term_tabelsize; /* CHANGE THIS */
  LEFT(res)  =  lft;
  KEY(res)   =  key;
  VALUE(res) =  val;
  RIGHT(res) =  rgt;
  return res;
}

/*
 *    Searching;
 *
 */
static lrc_term lookup(lrc_term key, AVL_tp avl, compare_function_tp comparer){
  int compare_res;

  if (avl==0) return 0;
  compare_res=comparer(key, KEY(avl));
  if (compare_res==lrc_less) return lookup(key, LEFT(avl),comparer);
  if (compare_res==lrc_equal) return VALUE(avl);
  return lookup(key, RIGHT(avl),comparer);
}

int avl_lookup(lrc_term * res, lrc_term key,
               AVL_tp avl, compare_function_tp comparer) {
  *res=lookup(key,avl,comparer);
  return (*res)!=0;
}

/* this is comment!!
AVL_tp avl_insert(lrc_term xk, lrc_term xv,
                  AVL_tp avl, compare_function_tp comparer){
  AVL_tp ins_right;
  int compare_res;

  if (avl==0) return create_bin(avl, xk, xv, avl);
  compare_res=comparer(xk, KEY(avl));
  if (compare_res==lrc_less) {
      AVL_tp ins_left = avl_insert(xk, xv, LEFT(avl), comparer);
      return  rebal(create_bin(ins_left, KEY(avl), VALUE(avl), RIGHT(avl)));
  }
  if (compare_res==lrc_equal) return create_bin(LEFT(avl), xk, xv, RIGHT(avl));
  ins_right = avl_insert(xk, xv, RIGHT(avl), comparer);
  return rebal(create_bin(LEFT(avl), KEY(avl), VALUE(avl), ins_right));
}
*/


/*
 *   Rotations
 */
INLINE static AVL_tp rotr(AVL_tp t){
  return create_bin(LEFT(LEFT(t)),KEY(LEFT(t)),VALUE(LEFT(t)),
                    create_bin(RIGHT(LEFT(t)),KEY(t),VALUE(t),RIGHT(t))
         );
}

INLINE static AVL_tp rotr_dot_create(AVL_tp lt, lrc_term kt, lrc_term vt, AVL_tp rt){
  return create_bin(LEFT(lt),KEY(lt),VALUE(lt),
                    create_bin(RIGHT(lt),kt,vt,rt)
         );
}

INLINE static AVL_tp rotl(AVL_tp t){
  return create_bin(
           create_bin(LEFT(t), KEY(t), VALUE(t), LEFT(RIGHT(t))),
           KEY(RIGHT(t)),
           VALUE(RIGHT(t)),
           RIGHT(RIGHT(t))
         );
}

/*   rotl_dot_create(lt,kt,vt,rt) == rotl(create_bin(lt,kt,vt,rt))
 */
INLINE static AVL_tp rotl_dot_create(AVL_tp lt, lrc_term kt, lrc_term vt, AVL_tp rt){
  return create_bin(
           create_bin(lt, kt, vt, LEFT(rt)),
           KEY(rt),
           VALUE(rt),
           RIGHT(rt)
         );
}

/*
 *    Shifts
 */

/* used by commented rebal
INLINE static AVL_tp shiftl(AVL_tp t){
  if (slope(RIGHT(t))==1)  {
      AVL_tp new_right = rotr(RIGHT(t));
      return rotl_dot_create(LEFT(t),KEY(t),VALUE(t),new_right);
  }
  return  rotl(t);
}
*/

INLINE static AVL_tp shiftl_dot_create(AVL_tp lt, lrc_term kt, lrc_term vt, AVL_tp rt){
  if (slope(rt)==1)  {
      AVL_tp new_right = rotr(rt);
      return rotl_dot_create(lt,kt,vt,new_right);
  }
  return  rotl_dot_create(lt,kt,vt,rt);
}

/* used by commented rebal
INLINE static AVL_tp shiftr(AVL_tp t) {
  if (slope(LEFT(t))==-1)  {
    AVL_tp new_left = rotl(LEFT(t));
    return rotr_dot_create(new_left,KEY(t),VALUE(t),RIGHT(t));
  }
  return rotr(t);
}
*/

INLINE static AVL_tp shiftr_dot_create(AVL_tp lt, lrc_term kt, lrc_term vt, AVL_tp rt) {
  if (slope(lt)==-1)  {
    AVL_tp new_left = rotl(lt);
    return rotr_dot_create(new_left,kt,vt,rt);
  }
  return rotr_dot_create(lt, kt, vt, rt);
}


/* 
**   Rebalancing
*/

/* used by commented insert
static AVL_tp rebal(AVL_tp t){
  int slope_t = slope(t);

  if (slope_t==2) return  shiftr(t);
  if (slope_t==-2) return  shiftl(t);
  return  t;
}
*/

INLINE static AVL_tp rebal_dot_create(AVL_tp lt, lrc_term kt, lrc_term vt, AVL_tp rt){
  int slope_t = fheight(lt)-fheight(rt);

  if (slope_t==2)  return shiftr_dot_create(lt, kt, vt, rt);
  if (slope_t==-2) return shiftl_dot_create(lt, kt, vt, rt);
  return  create_bin(lt, kt, vt, rt);
}

/*
**     Adding a key-value pair
*/

AVL_tp avl_insert(lrc_term xk, lrc_term xv,
                  AVL_tp avl, compare_function_tp comparer){
  AVL_tp ins_right;
  int compare_res;

  if (avl==0) return create_bin(avl, xk, xv, avl);
  compare_res=comparer(xk, KEY(avl));
  if (compare_res==lrc_less) {
      AVL_tp ins_left = avl_insert(xk, xv, LEFT(avl), comparer);
      if (ins_left==LEFT(avl)) return avl;
      return  rebal_dot_create(ins_left, KEY(avl), VALUE(avl), RIGHT(avl));
  }
  if (compare_res==lrc_equal) {
    if (xv==VALUE(avl)) return avl;
    return create_bin(LEFT(avl), xk, xv, RIGHT(avl));
  }
  ins_right = avl_insert(xk, xv, RIGHT(avl), comparer);
  if (ins_right==RIGHT(avl)) return avl;
  return rebal_dot_create(LEFT(avl), KEY(avl), VALUE(avl), ins_right);
}


/*
** Printing
*/
void avl_print(AVL_tp a, void print_key(), void print_value()) {
  if (a!=0) {
     printf("(");
     avl_print(LEFT(a),print_key,print_value);
     print_key(KEY(a)); printf("=");
     print_value(VALUE(a)); printf("\n");
     avl_print(RIGHT(a),print_key,print_value);
     printf(")");
  }
}

/* 
** MAARTEN PENNINGS started here
**
*/

#define AVL_IS_NIL(t)  (t==0)

/*
** Suppose we have the following tree. BIN nodes are marked with an astrix.
** They have three sons: a left son, a value son and a right son. Values
** are denoted by letters in this example. NIL trees are denoted by the
** caret.
**
**                   *
**                  /|\
**                 / m \
**                /     \
**               /       \
**              *         *
**             /|\       /|\
**            / d \     ^ w \
**           *     *         *
**          /|\   /|\       /|\
**         * b ^ ^ f *     ^ z ^
**        /|\       /|\
**       ^ a ^     ^ k ^
**
** Every * can be the current node. This is all we need to know of the tree!
** Note for people who know the history: the TOUCHED field is not needed,
** and we only use a handler to get back to the father O(1).
** 
** For example, in the figures below, we have marked the current node
** with a @ instead of a *. Every value that has already been retrieved
** is written in caps. As one can see, which letters should be caps
** and which lowercase can be deduced from the current node alone.
** All the values not exceeding the value belonging to the current node
** should be caps which is the same as all the values to the left of the @
** since AVL trees are ordered.
** 
**             *           :             *           :             *        
**            /|\          :            /|\          :            /|\       
**           / m \         :           / m \         :           / m \      
**          /     \        :          /     \        :          /     \     
**         /       \       :         /       \       :         /       \    
**        *         *      :        *         *      :        @         *   
**       /|\       /|\     :       /|\       /|\     :       /|\       /|\  
**      / d \     ^ w \    :      / d \     ^ w \    :      / D \     ^ w \ 
**     *     *         *   :     @     *         *   :     *     *         *
**    /|\   /|\       /|\  :    /|\   /|\       /|\  :    /|\   /|\       /|\ 
**   @ b ^ ^ f *     ^ z ^ :   * B ^ ^ f *     ^ z ^ :   * B ^ ^ f *     ^ z ^
**  /|\       /|\          :  /|\       /|\          :  /|\       /|\       
** ^ A ^     ^ k ^         : ^ A ^     ^ k ^         : ^ A ^     ^ k ^      
**  
**             *           :             *           :             @        
**            /|\          :            /|\          :            /|\       
**           / m \         :           / m \         :           / M \      
**          /     \        :          /     \        :          /     \     
**         /       \       :         /       \       :         /       \    
**        *         *      :        *         *      :        *         *   
**       /|\       /|\     :       /|\       /|\     :       /|\       /|\  
**      / D \     ^ w \    :      / D \     ^ w \    :      / D \     ^ w \ 
**     *     @         *   :     *     *         *   :     *     *         *
**    /|\   /|\       /|\  :    /|\   /|\       /|\  :    /|\   /|\       /|\ 
**   * B ^ ^ F *     ^ z ^ :   * B ^ ^ F @     ^ z ^ :   * B ^ ^ F *     ^ z ^
**  /|\       /|\          :  /|\       /|\          :  /|\       /|\       
** ^ A ^     ^ k ^         : ^ A ^     ^ K ^         : ^ A ^     ^ K ^      
**  
**             *           :             *           :             *   @?       
**            /|\          :            /|\          :            /|\          
**           / M \         :           / M \         :           / M \        
**          /     \        :          /     \        :          /     \       
**         /       \       :         /       \       :         /       \      
**        *         @      :        *         *      :        *         *     
**       /|\       /|\     :       /|\       /|\     :       /|\       /|\    
**      / D \     ^ W \    :      / D \     ^ W \    :      / D \     ^ W \   
**     *     *         *   :     *     *         @   :     *     *         *  
**    /|\   /|\       /|\  :    /|\   /|\       /|\  :    /|\   /|\       /|\ 
**   * B ^ ^ F *     ^ z ^ :   * B ^ ^ F *     ^ Z ^ :   * B ^ ^ F *     ^ Z ^
**  /|\       /|\          :  /|\       /|\          :  /|\       /|\       
** ^ A ^     ^ K ^         : ^ A ^     ^ K ^         : ^ A ^     ^ K ^      
**
** We model the retrieval of values from an AVL tree like a file.
** In the following table, the left colum gives our functions and the
** right colum the corresponding pascal file equivalent.
**
**   open_avl(handler,avl)                   reset(file,"external_name")
**   HEAD_KEY(handler), HEAD_TARGET(handle)  file^
**   EOAVL(handler)                          EOF(file)
**   get_avl(handler)                        get(file)
**
** Note that an AVL value corresponds in our case with a key->target pair.
** A file handler in pascal uses the notion of a head that moves over the
** file. For an AVL tree, the current node is this head.
** As noted above, it suffices to record the current node, but since it
** will take O(num nodes) to trace back the father of the current node, 
** we keep a stack of fathers.
*/

typedef struct avl_handler {
  struct avl_handler * next;
  AVL_tp  avl;
} * handler;

#define EOAVL(h)            (h==0)
#define HEAD(h)             ((h)->avl)
#define HEAD_KEY(h)         (KEY(HEAD(h)))
#define HEAD_TARGET(h)      (VALUE(HEAD(h)))


/* handler_pool is used to minimize calls to malloc and free.
   In handler_pool we store the freed handlers.
 */
#define HANDLER_POOL_SIZE 200

static handler handler_pool[HANDLER_POOL_SIZE];
static int     handler_pool_free=0;

static void free_handler(handler h) {
  if (handler_pool_free<HANDLER_POOL_SIZE) {
    handler_pool[handler_pool_free++]=h;
  }
  else FREE_AVL(h);
}

static handler alloc_handler() {
  if (handler_pool_free>0) {
    return handler_pool[--handler_pool_free];
  }
  else {
    handler res;
    res = (handler)ALLOC_AVL(sizeof(*res));
    return res;
  }
}

/*
** A handler, which is just a stack, has five standard procedures: 
** push and pop, HEAD (top) , EOAVL (is_empty) and
** empty (create_stack)). For debugging we provide a print routine.
*/

static void empty(handler *h) {
  *h=0;
}

static void push(handler *h, AVL_tp t) {
  handler res;
  res = (handler)alloc_handler(sizeof(*res));
  res->next=*h;
  res->avl=t;
  *h=res;
}

static void pop(handler *h) {
  handler hh;
  hh=(*h)->next;
  free_handler(*h);
  *h=hh;
}

/*
static void print_handler(handler h) {
  if (!EOAVL(h)) {
    avl_print(HEAD(h)); printf("\n");
    print_handler(h->next);
  }
}
*/

/*
** We will now implement an AVL-handler. We need a couple of
** functions on AVL trees, before we can implement the handler
** functions open_avl and get_avl.
*/

#define HEAD_LEFT(h)   (LEFT(HEAD(h)))
#define HEAD_RIGHT(h)  (RIGHT(HEAD(h)))

/*
** This function moves the current node or head push to the
** leftmost node. As a precondition we require for handle h
**        ! EOAVL(h)    &&   ! AVL_IS_NIL(HEAD(h))
*/

static void descend_leftmost(handler *h)  {
  while  (! AVL_IS_NIL(HEAD_LEFT(*h)) ) {
    push(h,HEAD_LEFT(*h));
    /* descend_leftmost(h); */
  }
}

static void open_avl(handler *h, AVL_tp t) {
  empty(h);
  if (!AVL_IS_NIL(t)) {
    push(h,t); descend_leftmost(h);
  }
}


/*
** The following function moves the current node or head to the next
** position. All the values to the left (including the values under the
** head) have been visited. Hence, we need to go to the right.
** Either the current node has a non-empty right subtree, whose
** leftmost node is the one we are looking for, or the right
** subtree is empty. Then, we need to go pop to the first node
** that has not been visited (read) yet, i.e. a node where we come in
** from the left.
*/

/* Precondition: ! EOAVL(h) */
static void goto_next_on_right(handler *h)  {
  if (AVL_IS_NIL(HEAD_RIGHT(*h))) { AVL_tp old_head;
    do {
      old_head=HEAD(*h); pop(h);
      if (EOAVL(*h)) return;
    } while ( HEAD_RIGHT(*h)==old_head );
  } else {
    push(h,HEAD_RIGHT(*h));
    descend_leftmost(h);
  }
}

#define get_avl(h) goto_next_on_right(h)

/*
** Comparing two AVL trees is easy now.
** We simply open_avl a handler for them, and compare element for
** element until one of the two files, ehh AVL trees, is empty.
*/


lrc_bool avl_equal(AVL_tp t1, AVL_tp t2
                  ,equal_function_tp eq_key
                  ,equal_function_tp eq_value) {
  handler h1, h2;
  lrc_bool equal=lrc_true;

  if (t1==t2) return lrc_true; /* just for speed */

  if (t1!=0 && t2!=0) {
    if (HASH(t1)!=HASH(t2)) return lrc_false;
  }

  open_avl(&h1,t1);
  open_avl(&h2,t2);
  while (equal && !EOAVL(h1) && !EOAVL(h2)) {
    equal = eq_key(HEAD_KEY(h1),HEAD_KEY(h2)) &&
            eq_value(HEAD_TARGET(h1),HEAD_TARGET(h2));
    get_avl(&h1); get_avl(&h2);
  }


  return equal && EOAVL(h1) && EOAVL(h2);
}


int avl_compare(AVL_tp t1,AVL_tp t2
               ,equal_function_tp eq_key   ,compare_function_tp cmp_key
               ,equal_function_tp eq_value ,compare_function_tp cmp_value) {
  handler h1, h2;
  lrc_bool equal=lrc_true;

  open_avl(&h1,t1);
  open_avl(&h2,t2);
  while (equal && !EOAVL(h1) && !EOAVL(h2)) {
    equal = eq_key(HEAD_KEY(h1),HEAD_KEY(h2)) &&
            eq_value(HEAD_TARGET(h1),HEAD_TARGET(h2));
    if (equal) {get_avl(&h1); get_avl(&h2); }
  }
  if ( EOAVL(h1)) {
    if ( EOAVL(h2)) {
      return lrc_equal;
    } else {
      return lrc_less;
    }
  } else {
    if ( EOAVL(h2)) {
      return lrc_greater;
    } else {
      int key_cmp = cmp_key(HEAD_KEY(h1),HEAD_KEY(h2));
      if (key_cmp!=lrc_equal) return key_cmp;
      return cmp_value(HEAD_TARGET(h1),HEAD_TARGET(h2));
    }
  }
}


/* insert all elements in t2 into t1
 *
 */
AVL_tp avl_update(AVL_tp t1, AVL_tp t2, compare_function_tp comparer) {
  handler h2;
  AVL_tp res=t1;

  if (t1==0) return t2;

  open_avl(&h2,t2);
  while (!EOAVL(h2)) {
     res=avl_insert(HEAD_KEY(h2),HEAD_TARGET(h2),res,comparer);
     get_avl(&h2);
  }
  return res;
}

/* Marking */
void mark_avl(AVL_tp t) {
  if (t) {
    mark_avl(LEFT(t));
    mark_term(VALUE(t));
    mark_term(KEY(t));
    mark_avl(RIGHT(t));
  }
}
