/*
** bools.c
** march 1994 by Maarten Pennings
** Library functions by Matthijs Kuiper
*/

#include "bools.h"

#define BOOL_INDEX_OFFSET 256
static lrc_term_ix index_bool(lrc_bool b) {
  /* lrc_bool is int so it can be negative (we took lrc_true==-1!) */
  return abs(b+T__Bool+BOOL_INDEX_OFFSET)%lrc_term_tabelsize;
}

static lrc_bool eq_bool(lrc_term t,lrc_ptr b) {
  return LRC_TAG(t)==FIELD(b,0,lrc_bool) && 
         LRC_VALUE(t,lrc_bool)==FIELD(b,sizeof(lrc_ptr),lrc_bool);
}

static lrc_term C_Bool_local(lrc_bool b) {
  struct {lrc_ptr tag; lrc_bool b;} buf;
  buf.tag=(lrc_ptr)T__Bool; buf.b=b;
  return lrc_shared_prim(index_bool(b),eq_bool,sizeof(buf),&buf);
}




static lrc_term true_term,false_term;

void lrc_Bool_init() {
  true_term =C_Bool_local(lrc_true);
  false_term=C_Bool_local(lrc_false);
  return;
}

void lrc_Bool_mark() {
  mark_term(true_term);
  mark_term(false_term);
  return;
}

lrc_term C_Bool(lrc_bool b) {
  lrc_term_register_hit(b ? true_term : false_term);
  return b ? true_term : false_term;
}


/*********************\
** Library functions **
\*********************/


int lrc_bool_compare(lrc_bool a, lrc_bool b) {
  if   (a) {if (b) return lrc_equal; else return lrc_greater; }
  else     {if (b) return lrc_less; else return lrc_equal; }
}

int lrc_bool_compare_term(lrc_term a, lrc_term b) {
  return lrc_bool_compare(LRC_VALUE(a,lrc_bool),LRC_VALUE(b,lrc_bool));
}

lrc_bool lrc_bool_equal(lrc_bool a, lrc_bool b) {
  /* a implies b AND b implies a == (!a || b) && (!b||a) */
  return (!a||b) && (!b||a);
}

lrc_bool lrc_bool_equal_term(lrc_term a, lrc_term b) {
  return lrc_bool_equal(LRC_VALUE(a,lrc_bool),LRC_VALUE(b,lrc_bool));
}

lrc_bool lrc_bool_default(void) { return lrc_true; }

