/*
** chars.c
** january 1994 by Maarten Pennings
** Library functions by Matthijs Kuiper
*/

#include <limits.h>
#include <stdio.h>
#include "chars.h"

static lrc_term_ix index_char(char c) {
  /* strange but lrc_true: chars can be negative */
  return abs(c+T__Char)%lrc_term_tabelsize;
}

static lrc_bool eq_char(lrc_term t,lrc_ptr b) {
  return LRC_TAG(t)==FIELD(b,0,int) && 
         LRC_VALUE(t,char)==FIELD(b,sizeof(lrc_ptr),char);
}

static lrc_term C_Char_local(char c) {
  struct {lrc_ptr tag;char c;} buf;
  lrc_term res;
  buf.tag=(lrc_ptr)T__Char; buf.c=c;
  res=lrc_shared_prim(index_char(c),eq_char,sizeof(buf),&buf);
  return res;
}



/* fast_neg stores [FAST_NEG,0) and
** fast_pos stores [0..FAST_POS)
*/
#if CHAR_MIN
#define FAST_NEG CHAR_MIN
#define FAST_POS CHAR_MAX
#else
#define FAST_NEG -1
#define FAST_POS CHAR_MAX
#endif
static lrc_term fast_neg[-(FAST_NEG)];
static lrc_term fast_pos[FAST_POS];

/* -Wchar-subscripts
**        Warn if an array subscript has type char.  This  is
**        a  common cause of error, as programmers often for-
**        get that this type is signed on some machines.
*/


void lrc_Char_init() {
  char c;
  /* 
  printf("Initializing characters.\n");
  printf("  FAST_NEG=%d\n", FAST_NEG);
  printf("  FAST_POS=%d\n", FAST_POS);
  */
  for(c=FAST_NEG;   c< 0; c++) {
     fast_neg[-c-1]=C_Char_local(c);
  }
  /* this loop does not terminate on sgi-mips...
  for(c=FAST_POS-1; c>=0; c--) {
     fast_pos[(int)c]=C_Char_local(c);
  }
  */
  for(c=FAST_POS-1; c>0; c--) {
     fast_pos[(int)c]=C_Char_local(c);
  }
  c=0; fast_pos[(int)c]=C_Char_local(c);
  return;
}

void lrc_Char_mark() {
  char c;
  for(c=FAST_NEG;   c< 0; c++) {
    mark_term(fast_neg[-c-1]);
  }
  for(c=FAST_POS-1; c>0; c--) {
    mark_term(fast_pos[(int)c]);
  }
  c=0; mark_term(fast_pos[(int)c]);
  return;
}

lrc_term C_Char(char c) {
  lrc_term res;
  if (c<FAST_NEG || c>=FAST_POS)
    res= C_Char_local(c);
  else { lrc_term t;
    if (c<0) {
      t=fast_neg[-c]; 
    }
    else {
      t=fast_pos[(int)c];
    }
    lrc_term_register_hit(t);
    res= t;
  }
  return res;
}

/*********************\
** Library functions **
\*********************/

int lrc_char_compare(char a, char b) {
    return (a<b) ? lrc_less : ((a==b)?lrc_equal:lrc_greater);
}

int lrc_char_compare_term(lrc_term a, lrc_term b) {
  return lrc_char_compare(LRC_VALUE(a,char),LRC_VALUE(b,char));
}

lrc_bool lrc_char_equal(char a, char b) {
  return a==b;
}

lrc_bool lrc_char_equal_term(lrc_term a, lrc_term b) {
  return lrc_char_equal(LRC_VALUE(a,char),LRC_VALUE(b,char));
}

void lrc_char_fprint(FILE * f, char c) {
  if (c=='\a') fprintf(f,"\\a"); else 
  if (c=='\n') fprintf(f,"\\n"); else 
  if (c=='\t') fprintf(f,"\\t"); else 
  if (c=='\v') fprintf(f,"\\v"); else 
  if (c=='\b') fprintf(f,"\\b"); else 
  if (c=='\r') fprintf(f,"\\r"); else 
  if (c=='\f') fprintf(f,"\\f"); else 
  if (c=='\'') fprintf(f,"\\'"); else 
  if (c=='\"') fprintf(f,"\\\""); else 
  if (c=='\\') fprintf(f,"\\\\"); else 
/*if (c=='\?') fprintf(f,"\\?"); else */
  if (c>=32 && c<=127) fprintf(f,"%c",c); else
  fprintf(f,"\\%o",(unsigned char)c);
}

void lrc_char_print(char c) {
  lrc_char_fprint(stdout,c);
}

char lrc_char_default(void) { return 'X'; }

