#ifndef FAT
#define FAT
#else
#  ifdef LEAN
   ___LEAN_IS_NOT_SUPPORTED___
#  endif
#endif
/****************************\
** Low level implementation **
\****************************/

/*
**  FAT implementation
**       0     1      2      3     4       5     6
**   +------+------+------+------+------+------+-------- ... --------+
**   | next | mark | hash | hits | birth| size | local info for cell |
**   +------+------+------+------+------+------+-------- ... --------+
**
**  LEAN implementation
**      0      1       2          3
**   +------+------+-----------+---------- ... --------+
**   | next | hash | mark+size |   local info for cell |
**   +------+------+-----------+---------- ... --------+
*/
#define ADMIN_SIZE      6

#define CELL_SIZE(s)    (ADMIN_SIZE*sizeof(lrc_ptr)+(s))
#define ADMIN(t,n,tp)   (*(((tp*)(t))+(n)))

typedef struct short_bits {
  unsigned short nnodes; unsigned rest:13, initial:1, mark:1, primitive:1; }
short_bits_tp;
#define SIZE(t)         ADMIN(t,ADMIN_SIZE-1,int)
#define BIRTH(t)        ADMIN(t,ADMIN_SIZE-2,int)
#define HITS(t)         ADMIN(t,ADMIN_SIZE-3,int)
#define HASH(t)         ADMIN(t,ADMIN_SIZE-4,lrc_term_ix)
#define MARK(t)         ADMIN(t,ADMIN_SIZE-5,short_bits_tp).mark
#define NNOD(t)         ADMIN(t,ADMIN_SIZE-5,short_bits_tp).nnodes
#define INITIAL(t)      ADMIN(t,ADMIN_SIZE-5,short_bits_tp).initial
#define PRIMITIVE(t)    ADMIN(t,ADMIN_SIZE-5,short_bits_tp).primitive
#define NEXT(t)         ADMIN(t,ADMIN_SIZE-6,lrc_ptr)

#define CLEAR_MARK(t) MARK(t)=0
#define SET_MARK(t) MARK(t)=1
#define CLEAR_INITIAL(t) INITIAL(t)=0
#define SET_INITIAL(t) INITIAL(t)=1
#define CLEAR_PRIMITIVE(t) PRIMITIVE(t)=0
#define SET_PRIMITIVE(t) PRIMITIVE(t)=1

/*
**   Layout of initial terms in memory: (FAT version only)
**
**       0       1       2         3          4         5     6
**   +--------+--------+--------+----------+----------+------+--- ... ----+
**   | parent | nnodes | shadow |    -     |   -      | size | local info |
**   |        | bits   |        |          |          |      |    ...     |
**   +--------+--------+--------+----------+----------+------+--- ... ----+
*/
#define SHADOW(t)       ADMIN(t,ADMIN_SIZE-4,lrc_ptr)
#define BITS(t)         ADMIN(t,ADMIN_SIZE-5,int)
#define PARENT(t)       ADMIN(t,ADMIN_SIZE-6,lrc_ptr)


