#include <stdio.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "ints.h"
#include "basics.h"
#include "term.h"
#include "grammar_tables.h"
#include "ebinding.h"

#define is_empty(lst) (LRC_TAG(lst)==T_Derived_BindingsNil)

#ifndef TRACE_HANDLING
#define TRACE_HANDLING 0
#endif

static binding_table_tp binding_table=0;
static int num_bindings=0;

void create_binding_table(int num_bind_constructs) {
  int ix;

  num_bindings=num_bind_constructs;
  if (num_bindings<=0) return;
  binding_table=(binding_table_tp)lrc_malloc(num_bindings * sizeof(*binding_table));
  for (ix=0;ix<num_bindings;ix++) {
    binding_table[ix].match_function=0;
    binding_table[ix].num_trafos=0;
    binding_table[ix].trafos=0;
  }
}

void define_binding(int bind_key, int num_trafos) {
  AssertI(bind_key<num_bindings,"define_binding:bind_key too large",bind_key);
  binding_table[bind_key].num_trafos=num_trafos;
  binding_table[bind_key].trafos= lrc_malloc(num_trafos * sizeof(trafo_entry_tp));
}

void set_binding_match_function(int bind_key, lrc_eb_match_function_tp match_function) {
  AssertI(bind_key<num_bindings,"set_binding_match_function:bind_key too large",bind_key);
  binding_table[bind_key].match_function=match_function;
}

void set_binding_trafo_function(int bind_key, int trafo_key,
                                lrc_eb_trafo_function_tp trafo_function) {
  AssertI(bind_key<num_bindings,"set_binding_trafo_function:bind_key too large",bind_key);
  AssertI(trafo_key<binding_table[bind_key].num_trafos,
          "set_binding_trafo_function:trafo_key too large",trafo_key);
  binding_table[bind_key].trafos[trafo_key].trafo_function=trafo_function;
}
int binding_num_trafos(int bind_key) {
  return binding_table[bind_key].num_trafos;
}

lrc_bool invoke_matcher(
   int bind_key, lrc_term event, lrc_term frame, lrc_term parameters) {

  AssertI(bind_key<num_bindings,"invoke_matcher:bind_key too large",bind_key);
  return binding_table[bind_key].match_function(event,frame,parameters);
}

lrc_term invoke_transformation(int bind_key, int trafo_key, 
                               lrc_term event, lrc_term frame, lrc_term parameters) {

  AssertI(bind_key<num_bindings,"invoke_transformation:bind_key too large",bind_key);
  AssertI(trafo_key<binding_table[bind_key].num_trafos,
          "invoke_transformation:trafo_key too large",trafo_key);
  return binding_table[bind_key].trafos[trafo_key].trafo_function(event,frame,parameters);
}

static int T__LrcBindingInstance,T_Derived_BindingsPair,T_Derived_BindingsNil;
static int T__LrcBindTrafos0;

static lrc_term add_binding_instance(
  lrc_term binding_instance, lrc_term instances) {

  return lrc_term2(T_Derived_BindingsPair,binding_instance,instances);
}

static lrc_term create_binding_instance(
     int binding_key, lrc_term match_parameters, lrc_term trafo_data) {
  return lrc_term3(T__LrcBindingInstance,C_Int(binding_key),match_parameters,trafo_data);
}

static int instance_matches(lrc_term event, lrc_term frame, lrc_term binding_instance) {
  int binding_number;
  lrc_term match_parameters;
  int res;

  binding_number=LRC_VALUE_int(LRC_CHILD(binding_instance,1));
  match_parameters=LRC_CHILD(binding_instance,2);
  if (TRACE_HANDLING) {
    printf("MATCHING binding_number=%d ->", binding_number); fflush(stdout);
  }
  if (TRACE_HANDLING) {
    printf("INVOKING MATCH FUNCTION of binding %d\n", binding_number); fflush(stdout);
    printf("   Event is "); lrc_term_print_long(event); printf("\n"); fflush(stdout);
    printf("   Frame is "); lrc_term_print_long(frame); printf("\n"); fflush(stdout);
    printf("   Match parameters are "); lrc_term_print_long(match_parameters); printf("\n"); fflush(stdout);
  }
  res=invoke_matcher(binding_number,event,frame,match_parameters);
  if (TRACE_HANDLING) {
    printf("MATCHING binding_number=%d ->", binding_number); fflush(stdout);
  }
  if (TRACE_HANDLING) {
    if (res) { printf(" YES"); } else { printf(" NO"); }
    printf("\n");
  }
  return res;
}

int is_empty_bindings(lrc_term instances) {
  return is_empty(instances);
}


lrc_term _lrc_add_bind(
   int binding_key, lrc_term match_parameters, lrc_term trafo_data, lrc_term instances) {
  /****
    printf("_lrc_add_bind:\n  binding_key=%d\n", binding_key);
    printf("  match_parameters @%p",match_parameters); lrc_term_print_long(match_parameters); printf("\n");
    printf("  trafo_data "); lrc_term_print_long(trafo_data); printf("\n\n");
  ****/
  return add_binding_instance(
           create_binding_instance(binding_key,match_parameters,trafo_data),
           instances);
}

lrc_term lrc_all_matches(lrc_term event, lrc_term frame, lrc_term instances) {
  lrc_term rest_matches;

  if (is_empty(instances)) return instances;
  rest_matches=lrc_all_matches(event,frame,LRC_CHILD(instances,2));
  if (instance_matches(event,frame,LRC_CHILD(instances,1))) {
    return lrc_term2(T_Derived_BindingsPair,LRC_CHILD(instances,1),rest_matches);
  }
  return rest_matches;
}

int instance_binding_number(lrc_term bi) {
  return LRC_VALUE_int(LRC_CHILD(bi,1));
}

lrc_term instance_match_parameters(lrc_term bi) {
  return LRC_CHILD(bi,2);
}

lrc_term instance_trafo_data(lrc_term bi) {
  return LRC_CHILD(bi,3);
}

lrc_term bind_trafo_subtree_reference(lrc_term bt) {
  return LRC_CHILD(bt,1);
}

lrc_term bind_trafo_parameters(lrc_term bt) {
  return LRC_CHILD(bt,2);
}

int is_empty_trafo_data(lrc_term td_list) {
  return LRC_TAG(td_list)==T__LrcBindTrafos0;
}


void init_event_bindings(void) {
  T__LrcBindingInstance=lrc_prod_name_to_tag("_LrcBindingInstance");
  T_Derived_BindingsPair=lrc_prod_name_to_tag("Derived_BindingsPair");
  T_Derived_BindingsNil=lrc_prod_name_to_tag("Derived_BindingsNil");
  T__LrcBindTrafos0=lrc_prod_name_to_tag("_LrcBindTrafos0");

}



