/*
** initial.c
** january 1996 by Dick Heijink
** may 1997, Matthijs Kuiper
*/

#include <memory.h>
#include <stdio.h>
#include "lrcassert.h"
#include "data_layout.h"
#include "basics.h"
#include "shalloc.h"
#include "ebinding.h"
#include "initial.h"

#include "grammar_tables.h"
#include "lrcmemory.h"

#ifndef TRACE_REPLACEMENTS
#define TRACE_REPLACEMENTS 0
#endif

extern void lrc_term_print_long(lrc_term);

static lrc_term initial_copy_node(lrc_term tree) {
  lrc_term copy;
  int size;
  
  Assert(lrc_is_initial(tree),"IN copy_initial_node:tree is not initial");

  size=CELL_SIZE(lrc_term_size(tree));
  copy=(lrc_term)lrc_malloc(size);
  copy=memcpy(copy,tree,size);
  return copy;
}

static lrc_term * lrc_copy_of_user_data(lrc_term t) {
   int data_size;
   lrc_term * copy;

   data_size=lrc_term_size(t);
   copy=(lrc_term *)lrc_malloc(data_size);
   if (TRACE_REPLACEMENTS) {
     printf("CREATING COPY of USER DATA\n");
     printf("DATA SIZE = %d.\n", data_size);
     { int ix;
       printf(" t[0]==%d\n", LRC_TAG(t));
       for(ix=1;ix<=lrc_constructor_num_sons(LRC_TAG(t)); ix++) {
         printf(" t[%d]==%d\n", ix, (int)LRC_CHILD(t,ix));
       }
     }
   }
   copy=memcpy(copy,LRC_USER_DATA(t),data_size);
   if (TRACE_REPLACEMENTS) {
     printf("END of COPY. COPY[0]=%d\n", (int)copy[0]);
     { int ix;
       for(ix=0;ix<=lrc_constructor_num_sons(LRC_TAG(t)); ix++) {
         printf(" copy[%d]==%d\n", ix, (int)copy[ix]);
       }
     }
   }

   return copy;
}

lrc_term shadow_replace_child(lrc_term t, int child_number, lrc_term replacement) {
  lrc_term parent;
  lrc_term shadow;
  lrc_term * node_copy;
  int size;
  parent=t;

  if (TRACE_REPLACEMENTS) {
    printf("REPLACING CHILD %d of\n", child_number);
    lrc_term_print_long(t); printf("\nby\n");
    lrc_term_print_long(replacement); printf("\n\n");
  }

  Assert(t!=0,"replace_child: t==0");
  Assert(replacement!=0,"shadow_replace_child: replacement==0");
  AssertI(child_number>0,"shadow_replace_child: child_number<=0", child_number);
  AssertI(child_number<=lrc_constructor_num_sons(LRC_TAG(t)),
          "shadow_replace_child: child_number too big", child_number);

  size=lrc_term_size(parent);
  node_copy=lrc_copy_of_user_data(parent);

  if (lrc_is_initial(replacement)) replacement=SHADOW(replacement);
  node_copy[child_number]=replacement;

 
  shadow=lrc_shared_cons(size,(lrc_term)node_copy);
  if (TRACE_REPLACEMENTS) {
    printf("REPLACEMENT="); lrc_term_print_long(shadow); printf("\n");
    Assert(!lrc_is_initial(shadow), "replace_child: result IS initial");
  }
  return shadow;
}
lrc_term initial_replace_child(lrc_term tree, int child_num, lrc_term replacement) {
  lrc_term res;

  Assert(lrc_is_initial(tree),"IN initial_replace_child: tree is not initial");
  Assert(lrc_is_initial(replacement),"IN initial_replace_child: replacement is not initial");

  res=initial_copy_node(tree);
  LRC_CHILD(res,child_num)=replacement;
  if (!PRIMITIVE(replacement)) PARENT(replacement)=res;
  SHADOW(res)=shadow_replace_child(SHADOW(tree),child_num,SHADOW(replacement));
  return res;
}

lrc_term lrc_create_initial(lrc_term shadow)
{
  lrc_term child;
  lrc_term initial;
  lrc_term_ix index;
  int size;
  int sons;
  int tag;

/**** 
  if (shadow==0) {
     printf("lrc_create_initial  shadow is <null>???");
     exit();
  }
  if ((int)shadow < 1000) {
     printf("scary shadow pointer in lrc_create_initial '%p'\n", shadow);
     exit();
  }
*****/
  if (lrc_is_initial(shadow)) {
     printf("lrc_create_initial:FAILURE: shadow is intial??\n");
     /*** lrc_term_print_long(shadow); printf("\n"); ***/
     return lrc_create_initial(SHADOW(shadow));
  }

  /* printf("lrc_create_initial: tag=%s\n", lrc_constructor_ident(LRC_TAG(shadow))); */

  /* printf("lrc_create_initial: tag=%d;size=%d\n", LRC_TAG(shadow), SIZE(shadow)); */
  /* 
  ** Retrieve the size (in bytes) of the local info for the cell.
  */
  size=SIZE(shadow);
  /*
  ** Allocate the memory required for the initial (topnode of the) term.
  */
  initial=lrc_malloc(CELL_SIZE(size));
  /*
  ** Initialise the internal ADMIN fields:
  ** - SIZE of initial term equals SIZE of shadow.
  ** - SHADOW of initial term equals the shadow (How trivial can you get?)
  ** - set initial-bit initial term
  ** - PARENT of initial term is unknown. (null-term)
  */
  SIZE(initial)=size;
  SHADOW(initial)=shadow;
  SET_INITIAL(initial);
  PARENT(initial)=TERM_NULL;
  if (PRIMITIVE(shadow)) SET_PRIMITIVE(initial);
  CLEAR_MARK(initial);
  /*
  ** Initialise the TAG field. A simple copy of the shadow term.
  */
  tag=LRC_TAG(initial)=LRC_TAG(shadow);
  /*
  ** Determine the number of sons of the term.
  */
  sons=lrc_constructor_num_sons(tag); 
  /*
  ** If the term does not have any sons, then the term is either a production
  ** without subterms or a primitive term.
  */
  if ( sons==0 ) {
    /*
    ** Copy the primitive value or do nothing in case of a structured term.
    */
     /* printf("Copying data (%d bytes); source=",size-sizeof(lrc_ptr));
        lrc_term_print_long(shadow); printf("\n");
      */
     memcpy(LRC_ADDRESS_PRIMITIVE(initial),LRC_ADDRESS_PRIMITIVE(shadow),
           size-sizeof(lrc_ptr));
     /*
       printf("  destination="); 
       lrc_term_print_long(shadow);printf("\n"); fflush(stdout);
      */
  }
  else { 
    /* 
    ** The shadow term is structured and contains one or more children.
    ** For every child :
    ** - create an initial term of the child
    ** - store this initial child in the parent node
    ** - set the parent link in the initial child
    */
    for (index=1; index <= sons; ++index ) {
      child=lrc_create_initial(LRC_CHILD(shadow,index));
      LRC_CHILD(initial,index)=child;
      PARENT(child)=initial;
    }
  }
  return initial;
}

lrc_term lrc_initial_parent(lrc_term term) {
  return (lrc_term)PARENT(term);
}

lrc_term lrc_initial_root(lrc_term term) {
  lrc_term t;
  Assert(lrc_is_initial(term),"IN lrc_initial_root: term is not initial");
  t=term;
  while (PARENT(t)!=TERM_NULL) t=lrc_initial_parent(t);
  return t;
}

lrc_term lrc_initial_shadow(lrc_term term) {
  return (lrc_term)SHADOW(term);
}

int lrc_is_primitive(lrc_term term) {
  return PRIMITIVE(term);
}


lrc_bool lrc_is_initial(lrc_term term) {
  return ( INITIAL(term) ? lrc_true : lrc_false );
}


lrc_term_ix lrc_initial_index_parent(lrc_term term) {
  lrc_term_ix index;
  lrc_term parent;
  int num_sons;
  
  parent=PARENT(term);
  num_sons=lrc_constructor_num_sons(LRC_TAG(parent));
  /*
  ** Search the children of the parent, until the given initial term
  ** is encountered.
  */
  for ( index=0; index<=num_sons; ++index )
    if ( LRC_CHILD(parent,index)==term )
      return index;
}

lrc_term lrc_shadow(lrc_term term) {
  /* If the term is INITIAL, then return the SHADOW.
  ** Else, simply return the term itself.
  */
  return (INITIAL(term) ? SHADOW(term) : term); 
}

