/*
** ints.c
** january 1994 by Maarten Pennings
** Library functions by Matthijs Kuiper
*/

#include <stdio.h>
#include "ints.h"

#define POSINT_INDEX_OFFSET 1000
#define NEGINT_INDEX_OFFSET 2000

#define LRC_ABS(a) ((a)<0?-(a):(a))
static lrc_term_ix index_int(int i) {
  return (i<0
          ? LRC_ABS(i+T__Int+NEGINT_INDEX_OFFSET)%lrc_term_tabelsize
          :  LRC_ABS(i-T__Int-POSINT_INDEX_OFFSET)%lrc_term_tabelsize);
}

static lrc_bool eq_int(lrc_term t,lrc_ptr b) {
  return LRC_TAG(t)==FIELD(b,0,int) && 
         LRC_VALUE(t,int)==FIELD(b,sizeof(lrc_ptr),int);
}

static lrc_term C_Int_local(int i) {
  struct {lrc_ptr tag;int i;} buf;
  buf.tag=(lrc_ptr)T__Int; buf.i=i;
  return lrc_shared_prim(index_int(i),eq_int,sizeof(buf),&buf);
}




/* fast_neg stores [FAST_NEG,0) and
** fast_pos stores [0..FAST_POS)
*/
#define FAST_NEG -400
#define FAST_POS 400
static lrc_term fast_neg[-(FAST_NEG)];
static lrc_term fast_pos[FAST_POS];

void lrc_Int_init() {
  int i;
  for(i=FAST_NEG;   i< 0; i++) fast_neg[-i-1]=C_Int_local(i);
  for(i=FAST_POS-1; i>=0; i--) fast_pos[i   ]=C_Int_local(i);
  return;
}

void lrc_Int_mark() {
  int i;
  for(i=FAST_NEG;   i< 0; i++) mark_term(fast_neg[-i-1]);
  for(i=FAST_POS-1; i>=0; i--) mark_term(fast_pos[i]);
  return;
}

lrc_term C_Int(int i) {
  if (i<FAST_NEG || i>=FAST_POS)
    return C_Int_local(i);
  else { lrc_term t;
    if (i<0) t=fast_neg[-i-1]; else t=fast_pos[i];
    lrc_term_register_hit(t);
    return t;
  }
}


/*********************\
** Library functions **
\*********************/

int lrc_int_compare(int a, int b) {
  return (a<b) ? lrc_less : ((a==b)?lrc_equal:lrc_greater);
}

int lrc_int_compare_term(lrc_term a, lrc_term b) {
  return lrc_int_compare(LRC_VALUE(a,int),LRC_VALUE(b,int));
}

lrc_bool lrc_int_equal(int a, int b) {
  return (a==b);
}

lrc_bool lrc_int_equal_term(lrc_term a, lrc_term b) {
  return lrc_int_equal(LRC_VALUE(a,int),LRC_VALUE(b,int));
}

int lrc_int_default(void) { return 0; }

