#include <stdio.h> 
#include <ctype.h>
#include <stdlib.h> 
#include <string.h>
#include <assert.h>
#include <lrclib.h>
#include "load_etf.h"

#define MAXLINE 400
#define MAX_ID_NAME 200
#define LENSUB 14
#define MAGIC_LINE "E#T#F#O#T#E#V#"

#define MAX_NUM_SONS 50
#define NUM_ITEMS 6
#define LIST_NUM_SONS 3
#define BINARY_NUM_SONS 2
#define NULLARY_NUM_SONS 0
#define FILLER '?'
#define FILLER_STRING "?"

#define map_ss_new(def) lrc_empty_map(lrc_string_equal_term,lrc_string_compare_term,lrc_string_equal_term,lrc_string_compare_term,C_Str(def))
#define map_it_new(def) lrc_empty_map(lrc_int_equal_term,lrc_int_compare_term,lrc_term_equal,LrcCompareTrees,C_Int(def))

/* opslag in ... */
static etf_structure final_etf;

static lrc_map strings;
static lrc_map categories;
static lrc_map category_properties;

static lrc_map op_name_to_etf_op_nr;
static lrc_map etf_op_nr_to_op_name;
static lrc_map etf_op_nr_to_gram_op_nr;

/* deze omvattende structuur is niet meer volgens de definitie (TO.doc)*/
struct struct_op
{
 lrc_map op_nr_num_children;
 lrc_map op_nr_num_attributes;
 lrc_map op_nr_is_primitive;
 lrc_map op_nr_category;
} node_properties;

static lrc_map node_nr_and_node;

lrc_map map_ss_set(lrc_map map, lrc_string string_key, lrc_string string_val)
{
 lrc_map result;
 result = lrc_map_single_update(C_Str(string_key), C_Str(string_val), map);
 return result;
}


lrc_map map_it_set(lrc_map map, int int_val, lrc_term term_key)
{
 lrc_map result;
 result = lrc_map_single_update(C_Int(int_val), term_key, map);
 return result;
}


lrc_string map_ss_apply(lrc_map map, lrc_string str_key)
{
 return LRC_VALUE_lrc_string(lrc_map_application(C_Str(str_key), map));
}


lrc_term map_it_apply(lrc_map map, int int_key)
{
 lrc_term res;
 res = lrc_map_application(C_Int(int_key), map);
 return res;
}

/*lrc_bool map_i_in (int key, lrc_map map);*/

lrc_bool map_i_in (int key, lrc_map map) {
 return lrc_map_in(C_Int(key), map);
}


int teller = 0;
char * line;
int token_counter;
char * token_ar[NUM_ITEMS];
int lineOK = 1;
char * retvalue;
static FILE * INFILE;
static int global_etf_op_nr;
static int in_string_section = 0;
static int vrije_string_ix;
static int vrije_category_ix;
static int vrije_op_ix; /* alle andere vrije_ variabelen ook globaal maken ??? */
static int vrije_node_ix;
static int operators_section = 0;
static int back_to_node;
static int etf_operator;
static int char_operator = -1;
static int str_operator = -1;
static int int_operator = -1;
static int bool_operator = -1;
static int real_operator = -1;
static lrc_string prim_value_str;
static char * binary_op_name;
static char * nullary_op_name;
static int binary_gram_op_nr;
static int nullary_gram_op_nr;

static void init_globals(void);

/************ M A I N  F U N C T I O N S **********/
static void do_magic_line(void);
static void do_tree_properties(void);
static void do_strings(void);
static void do_categories(void);
static void do_operators(void);
static lrc_term do_object(void);

/*************** OPEN, GET_LINE, SCAN_LINE ***************/
static int open_file(char * filename);
static void get_line(void);
static void scan_line(void);
static void scan_line_is_string(void);
static void line_is_tree_properties(void);
static void get_line_check_space(void);
static void set_token(int ix, char * string);
static void empty_token_ar(void);
static void check_token_ar(void);

/*************** OPERATORS & OBJECT SECTION etc. ***************/
static void enter_operator(void);
static lrc_term enter_term(void);
static lrc_term term1_2(void);
static lrc_term term3(void);
static lrc_term term4(void);
static lrc_term term5(void);
static void node_information(void);
static int process_operator_name(char * op_name);
static void fac_node_properties(int start_ix);
static lrc_term primitive_term(int k_etf_op_nr, char * string_repr);
static lrc_term get_term(int node_nr);
static lrc_term load_list(int keep_binary_gram_op_nr, int keep_nullary_gram_op_nr);
static void read_attributes(int keep_num_attributes);
static lrc_term build_term(int prod_tag, lrc_term *sons, int num_sons);
static lrc_term string_term(char * s);
static lrc_term int_term(char * s);
static lrc_term real_term(char * s);
static lrc_term char_term(char * s);
static lrc_term bool_term(char * s);

/*************** MISCELLANEOUS ***************/
static char * token_exists(int token_nr);
static int token_equal(int token_nr, char * token);
static void check_space_at_bol(void);
static void check_space_at_eol(void);
static void check_space_at_mid(void);
static void split_string(char * in_string, char ** s1, char ** s2, int start_at, char * splitchar);
static int operator_name_valid(char * name);
static int first_occurrence_in_etf(char * name);
static void node_properties_match_with_operators_section(int etf_op_nr, int num_attributes, int is_primitive);
static void lookup_etf_op_nr_in_operators_section(void);
static void lookup_string_nr_in_strings_section(int key);
static void compare_etf_with_grammar(int gram_op_nr, int num_children, int num_attributes, int is_primitive, char * category);
static void compare_num_children_with_list_num_children(char * list_op_name, int list_num_children);
static void convert_prim_val(char * prim_val);
static char convert_hex(char * part);


/*************** SETS & GETS ***************/
static void set_version(char * version);
static char * get_version();
static void set_tree_property(char * key, char * value);
static void get_tree_property();
static void enter_string(void);
static void set_string(char * s);
static char * get_string(int key);
static void enter_category(char * category, char * properties);
static void set_category(char * category);
static char * get_category(int key);
static void set_category_properties(char * category, char * properties);
static char * get_category_properties(char * category);
static void set_operator(char * op_name, int etf_op_nr, int gram_op_nr);
static void get_operator(); /* overbodig ??? */
static void set_node_properties(int etf_op_nr, int num_children, int num_attributes, int is_primitive, char * category);
static void get_node_properties(); /*overbodig*/

static void set_prim_etf_op_nr(int etf_op_nr);

static void set_etf_op_nr_to_gram_op_nr(int etf_op_nr, int gram_op_nr);
static int get_gram_op_nr_by_etf_op_nr(int etf_op_nr);
static void set_op_name_to_etf_op_nr(char * op_name, int etf_op_nr);
static int get_etf_op_nr_by_op_name(char * op_name);
static void set_etf_op_nr_to_op_name(int etf_op_nr, char * op_name);
static char * get_op_name_by_etf_op_nr(int etf_op_nr);
static void set_etf_op_nr_num_children(int etf_op_nr, int num_children);
static int get_etf_op_nr_num_children(int etf_op_nr);
static void set_etf_op_nr_num_attributes(int etf_op_nr, int num_attributes);
static int get_etf_op_nr_num_attributes(int etf_op_nr);
static void set_etf_op_nr_is_primitive(int etf_op_nr, int is_primitive);
static int get_etf_op_nr_is_primitive(int etf_op_nr);
static void set_etf_op_nr_category(int etf_op_nr, char * category);
static void get_etf_op_nr_category();

/*************** ERROR MESSAGES ***************/
static void enter_error(char * error_message);
static void undefined_gram_operator(char * op_name);
static void category_no_nonterminal(char * s1);
static void category_no_root(char * s1, char * s2);
static void name_no_identifier(char * this, char * op_name);
static void op_name_not_in_operators(char * op_name);
static void etf_op_nr_not_in_operators(int etf_op_nr);
static void num_children_not_correct(char * op_name, int right_num, int wrong_num);
static void num_children_not_conform_grammar(int etf_num_children, int gram_num_children);
static void category_not_in_categories_section(int cat_nr);
static void string_not_in_string_section(int str_nr);


/******************************************************************************************/
/* init_globals()                                                                         */
/* initializes variables and maps                                                         */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void init_globals() {
 int ix;

 for (ix = 0; ix < NUM_ITEMS; ix++) {
/*  token_ar[ix] = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);*/
  token_ar[ix] = (lrc_string)ALLOC_STR(MAXLINE);
 }
 line = (lrc_string)ALLOC_STR(MAXLINE);
 global_etf_op_nr = -1;
 binary_op_name = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);
 nullary_op_name = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);
 vrije_op_ix = 0;
/* prim_value_str = (lrc_string)ALLOC_STR(MAXLINE);*/
 final_etf.message = ALLOC_STR(MAXLINE);

 /* maps: new */
 final_etf.tree_properties = map_ss_new("0");
 strings = map_is_new("-1");
 categories = map_is_new("0");
 category_properties = map_ss_new("0");

 node_properties.op_nr_num_children = map_ii_new(-1);
 node_properties.op_nr_num_attributes = map_ii_new(-1);
 node_properties.op_nr_is_primitive = map_ii_new(0);
 node_properties.op_nr_category = map_is_new("-");

 etf_op_nr_to_gram_op_nr = map_ii_new(-1);
 op_name_to_etf_op_nr = map_si_new(-1);
 etf_op_nr_to_op_name = map_is_new("-");
 node_nr_and_node = map_it_new(-1);
}


static char * copy_string(char * s) {
 lrc_string scopy = (lrc_string)ALLOC_STR(strlen(s) + 1);
 return strcpy(scopy,s);
}



/************ M A I N  F U N C T I O N S **********/

/******************************************************************************************/
/* load_etf(char * filename)                                                              */
/* loads etf-file with name filename, and stores data in etf_structure final_etf          */
/* in : filename                                                                          */
/* out: etf_structure final_etf                                                           */
/******************************************************************************************/
etf_structure load_etf(char * filename) {
 lrc_term term;
 final_etf.status = 1;
 /* bij foutsituatie wordt deze waarde overschreven met 0. OK??? */

 init_globals();
 open_file(filename);
 if (final_etf.status == 0) {return final_etf;}

 /* start reading */
 get_line();
 if (final_etf.status == 0) {return final_etf;}
 if (!lineOK) {
  enter_error("Unexpected end of ETF-file, or empty line in ETF-file.\nMagic line not found on first line.");
  return final_etf;
 }

 do_magic_line();
 if (final_etf.status == 0) {return final_etf;}
 /*
 printf("\nTest version: \n");
 printf (final_etf.etf_version);
 printf("\n");
 */

 do_tree_properties();
 if (final_etf.status == 0) {return final_etf;}
 /*
 printf("\nTest properties:\n");
 lrc_map_print(final_etf.tree_properties,lrc_term_print_long,lrc_term_print_long);
 printf("\n");
 */

 do_strings();
 if (final_etf.status == 0) {return final_etf;}
/*
 printf("\nTest strings:\n");
 lrc_map_print(strings,lrc_term_print_long,lrc_term_print_long);
 printf("\n");
*/

 do_categories();
 if (final_etf.status == 0) {return final_etf;}
 /*
 printf("\nTest categories:\n");
 lrc_map_print(categories,lrc_term_print_long,lrc_term_print_long);
 lrc_map_print(category_properties,lrc_term_print_long,lrc_term_print_long);
 printf("\n");
 */

 do_operators();
 if (final_etf.status == 0) {return final_etf;}
/*
 printf("\nTest operators:\n");
 printf(":num_children:\n");
 lrc_map_print(node_properties.op_nr_num_children,lrc_term_print_long,lrc_term_print_long);
 printf("\n:num_attributes:\n");
 lrc_map_print(node_properties.op_nr_num_attributes,lrc_term_print_long,lrc_term_print_long);
 printf("\n:is_primitive:\n");
 lrc_map_print(node_properties.op_nr_is_primitive,lrc_term_print_long,lrc_term_print_long);
 printf("\n:category:\n");
 lrc_map_print(node_properties.op_nr_category,lrc_term_print_long,lrc_term_print_long);
 printf("\n\n");
*/

 term = do_object(); /* dat is: de (uiteindelijke) term */
 if (final_etf.status == 0) {return final_etf;}

/*lrc_map_print(node_nr_and_node,lrc_term_print_long,lrc_term_print_long);*/

 if (retvalue == 0) {
  printf("Einde ETF.\n");
 }
 else if (strlen(line) == 0) { /* lege regel, en dan EOF?? */
  get_line();
  if (retvalue == 0) {
   printf("Einde ETF.\n");
  }
  else {
   enter_error("Einde ETF (logisch einde) maar EOF van ETF-file nog niet bereikt.");
   return final_etf;
  }
 }
 else {
  enter_error("Einde ETF (logisch einde) maar EOF van ETF-file nog niet bereikt.");
  return final_etf;
 }
 
/* printf("\n--> uiteindelijke boom:\n"); lrc_term_print_long(term); printf("\n");*/
 if (final_etf.status == 0) {return final_etf;}
 final_etf.term = term; 
 if (final_etf.status == 1) {
  final_etf.message = NULL;
 }
return final_etf;
}



/************* M A I N  F U N C T I O N S ************/


/******************************************************************************************/
/* do_magic_line()                                                                        */
/* parses magic line, stores version                                                      */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void do_magic_line() {
 char * version;
/* ??? wat mag de lengte van de versie-string zijn?*/

 if (strncmp(token_ar[0],MAGIC_LINE,LENSUB) != 0) {
  enter_error("Magic line not conform specification.");
  return;
 }
 version = lrc_string_sub(token_ar[0], LENSUB+1, strlen(token_ar[0]));
 set_version(version);
 get_line();
 if (final_etf.status == 0) {return;}
 if (!lineOK) {
  enter_error("Unexpected end of ETF-file after Magic line.");
  return;
 }
}


/******************************************************************************************/
/* do_tree_properties()                                                                   */
/* parses tree properties section, stores tree propertey keys and values                  */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void do_tree_properties() {

 while ((lineOK) && token_equal(0, "#")) {
  /* de minimale properties-regel is #= */
  if (token_counter < 2) {
   enter_error("Invalid (empty) tree-properties-line.");
   return;
  }
  if (token_exists(2)) {
   if (!token_equal(2, "=")) { /* er zit geen =-teken in */
    enter_error("Tree properties-line lacks =-sign.");
    return;
   }
  }
  else {
   enter_error("Tree properties-line lacks =-sign.");
   return;
  }
  set_tree_property(token_ar[1], token_ar[3]);
  get_line();
  if (final_etf.status == 0) {return;}
 }

 if (!lineOK) {
  enter_error("Unexpected end of ETF-file after Magic line/at Tree properties section.");
  return;
 }
}


/******************************************************************************************/
/* do_string                                                                              */
/* parses string section, stores strings                                                  */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void do_strings() {
 vrije_string_ix = 0;

 if ((token_equal(0, "$")) &&
     (token_equal(1, "strings"))) {
  /* er is een $string-regel.*/
  in_string_section = 1;
  get_line();
  if (final_etf.status == 0) {return;}
  while ((lineOK) &&
         (in_string_section)) {
   enter_string();
   get_line(); /* hier is de mogelijkheid dat in_string_section op 0 wordt gezet */
   if (final_etf.status == 0) {return;}
  }
  if (!lineOK) {
   enter_error("Unexpected end of ETF-file at Strings section.");
   return;
  } 
  in_string_section = 0;
 }
}


/******************************************************************************************/
/* do_categories()                                                                        */
/* parses categories section, stores categories & properties                              */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void do_categories() {
 int nonterminal;

 vrije_category_ix = 0;
 if ((token_equal(0, "$")) &&
     (token_equal(1, "categories"))) {
  /* dit is de $categories section header */
  get_line();
  if (final_etf.status == 0) {return;}
  while ((lineOK) && (!token_equal(0, "$"))) {
   check_space_at_bol();
   if (final_etf.status == 0) {return;}
   check_space_at_eol();
   if (final_etf.status == 0) {return;}
   /* controleren of de category een nonterminal in de grammatica is */
   nonterminal = lrc_id_to_nont(token_ar[0]);
   if (nonterminal == -1) {
    category_no_nonterminal(token_ar[0]);
    return;
   }
   if (token_exists(1)) {
    /* hoe te controleren of de properties geldig zijn? ???*/
    /* kan niet, op de R=root na */
    if ((strchr(token_ar[1], 'R') != 0) &&
        (nonterminal != lrc_root_nonterminal())) {
    /* dus: er zit een R in de properties(-string) en
            de nonterminal-getallen van category en root komen niet overeen */
     category_no_root(token_ar[0], token_ar[1]);
     return;
    }
    enter_category(token_ar[0], token_ar[1]);
   }
   else {
    enter_category(token_ar[0], "");
   }
   get_line();
   if (final_etf.status == 0) {return;}
  }
  if (!lineOK) {
   enter_error("Unexpected end of ETF-file at Categories section.");
   return;
  } 
 }
}


/******************************************************************************************/
/* do_operators()                                                                         */
/* parses operator section, stores operator name & node properties                        */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void do_operators() {
 if (strcmp(line,"$operators") != 0) {
  enter_error("No $operators-line found, or ...\n... wrong lines preceeding $operators-line.");
  return;
 }
 get_line();
 if (final_etf.status == 0) {return;}
 while ((lineOK) && (strcmp(line,"$object") != 0)) {
  check_space_at_bol();
  if (final_etf.status == 0) {return;}
  check_space_at_eol();
  if (final_etf.status == 0) {return;}
  enter_operator();
  if (final_etf.status == 0) {return;}
  get_line();
  if (final_etf.status == 0) {return;}
 }
 if (!lineOK) {
   enter_error("Unexpected end of ETF-file at Operators section.");
   return;
 } 
}


/******************************************************************************************/
/* do_object()                                                                            */
/* parses object section, stores term                                                     */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static lrc_term do_object() {
 int num_nodes, num_strings;
 lrc_term term;

 if (strcmp(line,"$object") != 0) {
  enter_error("No $object-line found.");
  return term;
 }
 get_line_check_space();
 if (final_etf.status == 0) {return 0;}

 if (token_counter != 2) {
  enter_error("Line with number of nodes/number of strings not conform specification.");
  return 0;
 }
 /* printf("Aantal knopen is: %d.\nAantal strings is: %d.\n", num_nodes, num_strings);*/
 get_line_check_space();
 if (final_etf.status == 0) {return 0;}
 if (!lineOK) {
  enter_error("Unexpected end of ETF-file after num_nodes/num_strings in Object section.");
  return 0;
 } 
 vrije_node_ix = 0;
 term = enter_term();
 if (final_etf.status == 0) {return 0;}

/* Einde bestand bereikt.*/
 if (!lineOK) { /** ALLES GELEZEN **/
  /*
  printf("**************** Einde ETF-file. ****************\n");
  printf("\nTest object:\n");
  printf(":num_children:\n");
  lrc_map_print(node_properties.op_nr_num_children,lrc_term_print_long,lrc_term_print_long);
  printf("\n:num_attributes:\n");
  lrc_map_print(node_properties.op_nr_num_attributes,lrc_term_print_long,lrc_term_print_long);
  printf("\n:is_primitive:\n");
  lrc_map_print(node_properties.op_nr_is_primitive,lrc_term_print_long,lrc_term_print_long);
  printf("\n:category:\n");
  lrc_map_print(node_properties.op_nr_category,lrc_term_print_long,lrc_term_print_long);
  printf("\n\n");
  */
 }
 return term;
}


/*************** OPEN, GET_LINE, SCAN_LINE ***************/

/******************************************************************************************/
/* open_file(filename)                                                                    */
/* opens etf-file with filename                                                           */
/* in : filename                                                                          */
/* out: 0 if not opened, 1 if opened                                                      */
/******************************************************************************************/
static int open_file(char * filename){
 if (!(INFILE=fopen(filename,"r"))) {
  enter_error("File cannot be opened.");
  return 0;
 }
 return 1;
}


/******************************************************************************************/
/* get_line()                                                                             */
/* reads one line of input from the input-etf-file, calls scan_line for storing tokens    */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void get_line() {
 if (lineOK) {
  empty_token_ar();
  retvalue = fgets(line,MAXLINE,INFILE);
  if (retvalue == 0) {
   lineOK = 0;
   return;
  }
  if (line[strlen(line)-1] == '\n') {
  lineOK = 1;
  line[strlen(line)-1] = '\0';
  }
  if (line[0] == '\0') { /* lege regel */
   if (in_string_section == 0) { /* alleen mogelijk fout buiten de String section */
    retvalue = fgets(line,MAXLINE,INFILE);
    if (retvalue == 0) { /* einde van het bestand */
     lineOK = 0;
    }
    else { /* lege regel, niet aan het einde van het bestand */
      enter_error("Empty line encountered before ...");
    }
    return;
   }
  }
  if (in_string_section) {
   scan_line_is_string();
  }
  else {
   scan_line();
  }
  if (final_etf.status == 0) {return;}
  final_etf.etf_line = line;
 }
/* printf("Now! line %s lineok %i len %i.\n", line, lineOK, strlen(line));*/
}


/******************************************************************************************/
/* scan_line()                                                                            */
/* scans one line of input, stores tokens in token_ar                                     */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void scan_line() {
 int char_counter = 0;
 char * scan_pointer = line;

 if (scan_pointer[0] == '#') {
  line_is_tree_properties();
 }
 else {
  /* speciaal eerste character wegwerken */
  if (scan_pointer[0] == '!' ||
      scan_pointer[0] == '$' ||
      scan_pointer[0] == '.' ||
      scan_pointer[0] == '+' ||
      scan_pointer[0] == '-' ||
      scan_pointer[0] == '*') {
   token_ar[0][0] = scan_pointer[0];
   token_ar[0][1] = '\0';
   if (scan_pointer[0] == '+') {
    if (scan_pointer[1] == ' ') {
     scan_pointer++;
    }
    else if (scan_pointer[1] == '.') {
     if (!isdigit(scan_pointer[2])) {
      enter_error("Plus-sign and period not followed by number.");
      return;
     }
     token_counter++; /* volgende tokennummer */
     scan_pointer++; /* pointer verplaatsen */
     token_ar[token_counter][0] = '.'; /* punt in array */
     token_ar[token_counter][1] = '\0'; /* afsluiten */
    }
    else {
     enter_error("Plus-sign not followed by space or period.");
     return;
    }
   }
   else {
    if (scan_pointer[1] == ' ') {
     enter_error("Space after special (first) character not allowed.");
     return;
    }
   }
   scan_pointer++;
   token_counter++;
   char_counter = 0;
  }
  while (scan_pointer[0] != '\0') {
   token_ar[token_counter][char_counter] = scan_pointer[0];
   if ((scan_pointer[0] == ' ' ||
        scan_pointer[0] == '.') && 
       (scan_pointer[1] != ' ')) {
    if (scan_pointer[0] == '.') {
     token_ar[token_counter][char_counter] = '\0'; /* huidige token afsluiten */
     token_counter++; /* volgende tokennummer */
     char_counter = 0;
     token_ar[token_counter][char_counter] = '.'; /* punt in array */
     char_counter++;
    }
    token_ar[token_counter][char_counter] = '\0';
    token_counter++;
    char_counter = 0;
   }
   else {
    char_counter++;
   }
   scan_pointer++;
   if (token_counter > 5) {
    enter_error("Line contains more than 6 tokens."); /*token_counter begint bij 0 */
    return;
   }
  }
  token_ar[token_counter][char_counter] = '\0';
  token_counter++;
 }
 check_token_ar();
}


/******************************************************************************************/
/* scan_line_is_string()                                                                  */
/* scans one line of input, stores token (=string, = line) in token_ar[0] of token_ar     */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void scan_line_is_string() {
 if ((strcmp(line, "$categories") == 0) ||
     (strcmp(line, "$operators") == 0)) {
  in_string_section = 0;
  set_token(0, "$");
  if (strcmp(line, "$categories") == 0) {
   set_token(1, "categories");
  }
  else {
   set_token(1, "operators");
  }
  token_counter = 2;
 }
 else {
  set_token(0, line);
  token_counter = 1;
 }
 check_token_ar();
}


static void line_is_tree_properties() {
 token_ar[0][0] = '#';
 token_ar[0][1] = '\0';
 split_string(line, &token_ar[1], &token_ar[3],  2, "=");
 /* als het split-teken niet bestaat, dan gaat line in het eerste token */
 /* token_ar[2] is nog steeds ?\0 */
 /* token_ar[3] is een lege string */
 /*  printf("Key is %s, value is %s.\n", token_ar[1], token_ar[3]);*/
 if (strlen(token_ar[1]) == 0) { /* de key-string is leeg;
                                    het is dan #
                                            of #=
                                            of #=xxxx*/
  if (strcmp(line, "#") == 0) { /* # */
   token_counter = 1;
   token_ar[1][0] = FILLER;
   token_ar[1][1] = '\0';
   token_ar[3][0] = FILLER;
   token_ar[3][1] = '\0';
  }
  else { /* key is leeg, value is al dan niet leeg */
   strcpy(token_ar[2], "=");
   token_counter = 4;
  }
 }
 else if (strlen(token_ar[3]) == 0) { /* de value-string is leeg */
  if (line[strlen(line)-1] == '=') {
   strcpy(token_ar[2], "=");
   token_counter = 4;
  }
  else {
   token_counter = 2;
   token_ar[3][0] = FILLER;
   token_ar[3][1] = '\0';
  }
 }
 else {
  strcpy(token_ar[2], "=");
  token_counter = 4;
 }
}


static void set_token(int ix, char * string) {
  strcpy(token_ar[ix], string);
}

static void empty_token_ar() {
 int ix;

 token_counter = 0;
 for (ix = 0; ix < NUM_ITEMS; ix++) {
  token_ar[ix][0] = FILLER;
  token_ar[ix][1] = '\0';
 }
}


static void check_token_ar() {
 int ix;

 if (token_counter == 0) {
  enter_error("No tokens counted.");
  return;
 }

 for (ix = 0; ix < NUM_ITEMS; ix++) {
  if (ix < token_counter) { /* deze elementen moeten een waarde bevatten */
   if (in_string_section == 0) { /* d.w.z. buiten de String section */
    if (token_ar[ix][0] == FILLER && token_ar[ix][1] == '\0') {
/*     printf("index %d, token-counter %d, contents %s\n", ix, token_counter, token_ar[ix]);*/
     enter_error("Element of token-array has no contents.");
     return;
    }
   }
  }
  else { /* deze elementen moeten leeg zijn */
   if (!(token_ar[ix][0] == FILLER && token_ar[ix][1] == '\0')) {
/*    printf("index %d, token-counter %d, contents %s\n", ix, token_counter, token_ar[ix]);*/
    enter_error("Element of token-array should be empty.");
    return;
   }
  }
 }
}


static void get_line_check_space() {
 get_line();
 if (final_etf.status == 0) {return;}
 check_space_at_bol();
 if (final_etf.status == 0) {return;}
 check_space_at_eol();
 if (final_etf.status == 0) {return;}
 check_space_at_mid();
 if (final_etf.status == 0) {return;}
}


/*************** OPERATORS & OBJECT SECTION etc. ***************/
static void enter_operator() {
 int etf_op_nr = vrije_op_ix;
 char * op_name;
 int num_children;
 int num_attributes;
 int is_primitive;
 char * category = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);

 operators_section = 1;
 strcpy(category,"-");
 if (token_counter < 4) {
  enter_error("Composition of operator-line or object-line is not correct.");
  return;
 }
 if (is_primitive == 1 && num_children != 0) {
  enter_error("Primitive = 1, but number of children <> 0.");
  return;
 }
/* strcpy(op_name, token_ar[0]); kan niet ... */
 op_name = copy_string(token_ar[0]);
 num_children = atoi(token_ar[1]);
 num_attributes = atoi(token_ar[2]);
 is_primitive = atoi(token_ar[3]);
 if (token_exists(4)) {
  category = copy_string(token_ar[4]);
 }

 /* t.b.v. operatorregels moet er gecheckt worden of name er al is.
    operator-regels moeten uniek zijn.
 */   
/*printf("voor uni\n");
 lrc_map_print(op_name_to_etf_op_nr,lrc_term_print_long,lrc_term_print_long);*/
 if (first_occurrence_in_etf(op_name)) {
  int gram_op_nr;
 /* controleren of name geldig is binnen de grammatica */
  gram_op_nr = operator_name_valid(op_name);
  if (final_etf.status == 0) {return;}
  set_operator(op_name, etf_op_nr, gram_op_nr);
  /* bewaren van etf_op_nr van de primitieven (in primitief-soort-veld), 
     om bij het vullen van de term te weten om wat voor soort primval het gaat */
  /* printf("primi = %d\n", is_primitive);*/
  if (is_primitive) {
   set_prim_etf_op_nr(etf_op_nr);
   if (final_etf.status == 0) {return;}
  }
  global_etf_op_nr = etf_op_nr;
  /* i.t.t. etf-perl moet deze test hier ook tussen. E.e.a. moet altijd equal zijn aan de gramm, uniek of niet.*/
  compare_etf_with_grammar(get_gram_op_nr_by_etf_op_nr(etf_op_nr), num_children, num_attributes, is_primitive, category);
  if (final_etf.status == 0) {return;}
  set_node_properties(etf_op_nr, num_children, num_attributes, is_primitive, category);
 }
 else {
  enter_error("Operator-name is not unique.");
  return;
 }
 operators_section = 0;
}


static lrc_term enter_term() {
 lrc_term node;

 if (!lineOK) {
  enter_error("Unexpected end of ETF-file, object section.");
  return 0;
 }
 /* TERM */
 /******** 1 (term) ********/
 /******** 2 (primitive term) ********/
 /********************************************************************************/
 /* term : node-information  NL  term*  attribute*                               */
 /*      | node-information  NL  +  SPACE  string  NL                            */
 /********************************************************************************/
 if (token_equal(0, "!") ||
     isdigit(token_ar[0][0])) {
  node = term1_2();
  if (final_etf.status == 0) {return 0;}
 }
 /******** 3 (lijst) ********/
 /********************************************************************************/
 /* term : *binary  SPACE  nullary  NL  term*  nullary  NL                       */
 /********************************************************************************/
 else if (token_equal(0, "*")) {
  node = term3();
  if (final_etf.status == 0) {return 0;}
 }
 /******** 4 (pointer naar voorgaande knoop) ********/
 /********************************************************************************/
 /* term : -number  NL                                                           */
 /********************************************************************************/
 else if (token_equal(0, "-")) {
  node = term4();
  if (final_etf.status == 0) {return 0;}
 }
 /******** 5 (base64number) ********/
 /********************************************************************************/
 /* term : base64number  NL                                                      */
 /********************************************************************************/
 else {
  node = term5();
  if (final_etf.status == 0) {return 0;}
 }
 return node;
}


static lrc_term term1_2() {
 lrc_term node;
 int keep_etf_op_nr;
 int keep_num_attributes;
 int node_ix;

 node_information();
 if (final_etf.status == 0) {return 0;}
 keep_etf_op_nr = global_etf_op_nr;
 keep_num_attributes = get_etf_op_nr_num_attributes(global_etf_op_nr);

 get_line_check_space();
 if (final_etf.status == 0) {return 0;}

 if (token_exists(0)) {
  if (token_equal(0, "+")) {
   if (token_exists(1)) {
    if (token_equal(1, ".")) { /* het is .number uit string section */
     if (token_exists(2)) {
      lookup_string_nr_in_strings_section(atoi(token_ar[2]));
      if (final_etf.status == 0) {return 0;}
      prim_value_str = get_string(atoi(token_ar[2]));
     }
    }
    else { /* de tussenliggende spatie is niet in token_ar opgeslagen. */
     prim_value_str = copy_string(token_ar[1]);
     /* er wordt niet gekeken of dit een str of een int is */
    }
    /* is deze test zo voldoende??? */
    if (global_etf_op_nr == str_operator) {
/*     convert_prim_val(prim_value_str);  deze functie tijdelijk niet gebruiken */
     if (final_etf.status == 0) {return 0;}
    }
   }
  }
 }

 if (get_etf_op_nr_is_primitive(keep_etf_op_nr)) {
  if (final_etf.status == 0) {return 0;}
  node = primitive_term(keep_etf_op_nr, prim_value_str);
  /* in deze functie kan 'niets' verkeerd gaan, dus status wordt er niet 0 */
  get_line_check_space();
  if (final_etf.status == 0) {return 0;}
  return node;
 }
 else {
  int t_num_children;
  int prod_tag;

 node_ix = vrije_node_ix;
 vrije_node_ix++; /* hij hoogt vaker op dan dat er build_terms zijn ??? */
/* printf("vni %d\n", vrije_node_ix);*/

  t_num_children = get_etf_op_nr_num_children(keep_etf_op_nr);
  if (t_num_children > 0) {
   lrc_term son_table[MAX_NUM_SONS];
   int ix;
/*printf ("hoeveel kids %d van : %d\n", t_num_children, get_gram_op_nr_by_etf_op_nr(keep_etf_op_nr));*/
   for (ix = 1; ix <= t_num_children; ix++) {
    son_table[ix] = enter_term();
    if (final_etf.status == 0) {return 0;}
   }
   if (keep_num_attributes > 0) {
    read_attributes(keep_num_attributes);
   }
   if (final_etf.status == 0) {return 0;}
   prod_tag = get_gram_op_nr_by_etf_op_nr(keep_etf_op_nr);
   /* prod_tag is op_nr van de grammatica */
   node = build_term(prod_tag, son_table, t_num_children);
   /* in deze functie kan 'niets' verkeerd gaan, dus status wordt er niet 0 */
  }
  else {
   lrc_term son_table[1];
   read_attributes(keep_num_attributes);
   prod_tag = get_gram_op_nr_by_etf_op_nr(keep_etf_op_nr);
   node = build_term(prod_tag, son_table, 0);
  }
 }
 node_nr_and_node = map_it_set(node_nr_and_node, node_ix, node);
 return node;
}


static lrc_term term3() { /* lijst */
 lrc_term list_node;
 int gram_op_nr;
 int keep_binary_gram_op_nr;
 int keep_nullary_gram_op_nr;
 int etf_op_nr = vrije_op_ix;
 char * category = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);
 int first_occurrence;

 strcpy(category,"-");
 /* de binary */
 if (isdigit(token_ar[1][0])) { /* een nummer */
  global_etf_op_nr = atoi(token_ar[1]);
  lookup_etf_op_nr_in_operators_section();
  if (final_etf.status == 0) {return 0;}
  binary_gram_op_nr = get_gram_op_nr_by_etf_op_nr(global_etf_op_nr);
  keep_binary_gram_op_nr = binary_gram_op_nr;
 }
 else {
  binary_op_name = copy_string(token_ar[1]);
  first_occurrence = process_operator_name(binary_op_name); /*slaat gegevens op, retourneert 1 als 'eerste voorkomen' */
  if (final_etf.status == 0) {return 0;}
  compare_num_children_with_list_num_children(binary_op_name, BINARY_NUM_SONS);
  if (final_etf.status == 0) {return 0;}
  if (first_occurrence) {
   set_node_properties(etf_op_nr, BINARY_NUM_SONS, 0, 0, category);
  }
  keep_binary_gram_op_nr = lrc_prod_name_to_tag(binary_op_name);
 }
 /* de nullary */
 if (isdigit(token_ar[2][0])) { /* een nummer */
  global_etf_op_nr = atoi(token_ar[2]);
  lookup_etf_op_nr_in_operators_section();
  if (final_etf.status == 0) {return 0;}
  nullary_gram_op_nr = get_gram_op_nr_by_etf_op_nr(global_etf_op_nr);
  keep_nullary_gram_op_nr = nullary_gram_op_nr;
 }
 else {
  nullary_op_name = copy_string(token_ar[2]);
  first_occurrence = process_operator_name(nullary_op_name);
  if (final_etf.status == 0) {return 0;}
  compare_num_children_with_list_num_children(nullary_op_name, NULLARY_NUM_SONS);
  if (final_etf.status == 0) {return 0;}
  if (first_occurrence) {
   set_node_properties(etf_op_nr, 0, 0, 0, category);
  }
  keep_nullary_gram_op_nr = lrc_prod_name_to_tag(nullary_op_name);
 }
 /*printf("het nummer is %d van naam %s\n", keep_nullary_gram_op_nr, nullary_op_name);*/
 get_line_check_space();
 if (final_etf.status == 0) {return 0;}
 list_node = load_list(keep_binary_gram_op_nr, keep_nullary_gram_op_nr);
 if (final_etf.status == 0) {return 0;} /* overbodig ??? NEE */
 return list_node;
}


static lrc_term term4() { /* pointer naar voorgaande knoop */
 lrc_term node;
 int node_nr = atoi(token_ar[1]);

 back_to_node = (vrije_node_ix - node_nr);
 /*
 printf("Pointer -%s; node_nr %d, vrije_node_ix %d; back_to_node %d.\n", token_ar[1], node_nr, vrije_node_ix ,back_to_node);
 */
 get_line_check_space();
 if (final_etf.status == 0) {return 0;}
 return get_term(back_to_node);
}

static lrc_term term5() { /* base64number */
 lrc_term node;
 enter_error("Verwacht: base64number. Not yet implemented.");
 return 0;
}


static void node_information() {
 char * op_name;

 if (token_equal(0, "!")) { /* node-information begint met ! */
  op_name = copy_string(token_ar[1]);
  process_operator_name(op_name);
  if (final_etf.status == 0) {return;}
  if (token_counter >= 3) {
   fac_node_properties(2); /* 2 is startindex */
   if (final_etf.status == 0) {return;}
  }
 }
 else if (isdigit(token_ar[0][0])) { /* node-information is een number */
  global_etf_op_nr = atoi(token_ar[0]);
  lookup_etf_op_nr_in_operators_section();
  if (final_etf.status == 0) {return;}
  if (token_counter >= 2) {
   fac_node_properties(1); /* 1 is startindex */
  }
 }
 else {
  enter_error("Node_information is not correct.");
  return;
 }
}


static int process_operator_name(char * op_name) {
 int gram_op_nr;

 if (first_occurrence_in_etf(op_name)) {
  /* controleren of name geldig is binnen de grammatica */
  gram_op_nr = operator_name_valid(op_name);
  if (final_etf.status == 0) {return 0;}
  global_etf_op_nr = vrije_op_ix;
  set_operator(op_name, global_etf_op_nr, gram_op_nr);
 }
 else {
  /* niets te controleren? */
  global_etf_op_nr = get_etf_op_nr_by_op_name(op_name);
 }
 return first_occurrence_in_etf(op_name);
}


static void fac_node_properties(int start_ix) {
 int num_children;
 int num_attributes;
 int is_primitive;
 char * category = (lrc_string)ALLOC_STR(MAX_ID_NAME + 1);
 int prop_counter = token_counter - start_ix;

 num_attributes = -1;
 is_primitive = -1;
 strcpy(category,"-");
 num_children = atoi(token_ar[start_ix]);
 if (prop_counter < 2 || prop_counter > 4) {
   enter_error("Number of elements on object line not correct.");
   return;
 }
 if (prop_counter == 2) { /* num_children en category */
   category = copy_string(token_ar[start_ix + 1]);
 }
 if (prop_counter >= 3) { /* num_children, num_attributes, is_primitive */
  num_attributes = atoi(token_ar[start_ix + 1]);
  is_primitive = atoi(token_ar[start_ix + 2]);
 }
 if (prop_counter == 4) { /* num_children, num_attributes, is_primitive, category */
   category = copy_string(token_ar[start_ix + 3]);
 }
 compare_etf_with_grammar(get_gram_op_nr_by_etf_op_nr(global_etf_op_nr), num_children, num_attributes, is_primitive, category);
 if (final_etf.status == 0) {return;}
 /* attr en prim kunnen niet met de gramm gecheckt worden, dus dan maar met de operators section */
  node_properties_match_with_operators_section(global_etf_op_nr, num_attributes, is_primitive);
 if (final_etf.status == 0) {return;}
 set_node_properties(global_etf_op_nr, num_children, num_attributes, is_primitive, category);

}


static lrc_term get_term () {
 return map_it_apply(node_nr_and_node, back_to_node);
}


static lrc_term primitive_term(int k_etf_op_nr, char * string_repr) {
 lrc_term node;

 if (k_etf_op_nr == int_operator) {
  node = int_term(string_repr);
 }
 else if (k_etf_op_nr == char_operator) {
  node = char_term(string_repr);
 }
 else if (k_etf_op_nr == bool_operator) {
  node = bool_term(string_repr);
 }
 else if (k_etf_op_nr == real_operator) {
  node = real_term(string_repr);
 }
 else {
  node = string_term(string_repr);
 }
 return node;
}


static lrc_term load_list(int keep_binary_gram_op_nr, int keep_nullary_gram_op_nr) {
 lrc_term son_table[LIST_NUM_SONS];
 int num_sons;

 son_table[1] = enter_term();
/* printf("wat is er mis? status is %d %s.\n", final_etf.status, final_etf.message);*/
/*printf("--> na enter_tree in loadlist "); lrc_term_print_long(son_table[1]); printf("\n");*/

 if (final_etf.status == 0) {return 0;}

/* printf("1 lrc tag %d keep_nul %d\n", LRC_TAG(son_table[1]), keep_nullary_gram_op_nr);*/

 if (LRC_TAG(son_table[1]) == keep_nullary_gram_op_nr) {
  return son_table[1];
 }

 vrije_node_ix++;
/* printf("vni %d\n", vrije_node_ix);*/

 son_table[2] = load_list(keep_binary_gram_op_nr, keep_nullary_gram_op_nr);
 if (final_etf.status == 0) {return 0;} /* deze moet hier wel staan!!! */

 num_sons = lrc_constructor_num_sons(keep_binary_gram_op_nr); /* altijd 2 ??? */
 return build_term(keep_binary_gram_op_nr, son_table, num_sons);
}


static void read_attributes(int keep_num_attributes) {
 int i;

 if (!lineOK) {
  enter_error("Unexpected end of ETF-file, no attributes to read.");
  return;
 }
 for (i = 1; i <= keep_num_attributes; i++) {
 /* er komt eerst een attribute name, d.w.z. een .identifier of een .number */
  if (!token_equal(0, ".")) {
   enter_error("Line with attribute-name does not start with a period.");
   return;
  }
  if (isdigit(token_ar[1][0]) || /* een nummer */
      isalpha(token_ar[1][0])) { /* letter */ /* ??? zeer beperkte test */ 
   /* nu komt er -ofwel een attribute-value,
                 -ofwel een nieuwe regel, en dan een term */
   if (isdigit(token_ar[1][0])) {
    lookup_string_nr_in_strings_section(atoi(token_ar[1]));
    if (final_etf.status == 0) {return;}
   }
   if (token_equal(2, "")) { /* er is niets meer op deze regel */
    get_line_check_space();
    if (!lineOK) {
     enter_error("Unexpected end of ETF-file, while reading attributes (no term following attribute-name-line).");
     return;
    }
    if (final_etf.status == 0) {return;}
    enter_term();
    if (final_etf.status == 0) {return;}
   } 
   else { /* dan moet er een attribute-value komen, dat is spatie-string of .number */
          /* de spatie wordt niet in token-ar opgenomen ... */
    if (token_equal(2, ".")) { /* punt */
     if (isdigit(token_ar[3][0])) { /* een nummer, alles in orde?, volgende regel */
      lookup_string_nr_in_strings_section(atoi(token_ar[3]));
      if (final_etf.status == 0) {return;}
      get_line_check_space();
      if (final_etf.status == 0) {return;}
     }
     else {
      enter_error("Attribute-value: .number expected.");
      return;
     }
    }
    else { /* string, alles in orde, volgende regel*/ /* ??? alles wordt geaccepteerd */
     get_line_check_space();
     if (final_etf.status == 0) {return;}
    }
   }
  }
  else {
   enter_error("Line with attribute-name (.identifier or .number) not correct.");
   return;
  }
 }
 return;
}


static lrc_term build_term(int prod_tag, lrc_term *sons, int num_sons) {
 lrc_term node;
 teller ++;
 /* printf("hierlangs %d\n", teller);*/
 /* printf("building with production %d\n", prod_tag);*/
 sons[0] = (lrc_ptr)prod_tag;
 node = lrc_shared_cons((num_sons+1)*sizeof(sons[0]),(lrc_term)sons);
 /* printf("--> "); lrc_term_print_long(node); printf("\n");*/
 return node;
}


static lrc_term string_term(char * s) {
 lrc_term node;
 node = C_Str(copy_string(s));
 return node;
}


static lrc_term int_term(char * s) {
 lrc_term node;
 int i;
/* sscanf(s,"%d",&i); dit sloopt + en voorloopnullen eraf ... */
 sscanf(s,"%d",&i);
 node = C_Int(i);
 return node;
}


static lrc_term real_term(char * s) {
 lrc_term node;
 double r;
 r = STRtoREAL(s);
 node = C_Real(r);
 return node;
}


static lrc_term char_term(char * s) {
 lrc_term node;
 int i;
 sscanf(s,"%d",&i);
 node = C_Char((char)i);
 return node;
}


static lrc_term bool_term(char * s) {
 lrc_term node;
 int i;
 sscanf(s,"%d",&i);
 node = C_Bool(i==0?lrc_false:lrc_true);
 return node;
}


/*************** MISCELLANEOUS ***************/
static char * token_exists(int token_nr) {
 if ((token_nr + 1) <= token_counter) {
  return token_ar[token_nr];
 }
 else {
  return NULL;
 }
}


static int token_equal(int token_nr, char * token) {
/*printf("token: nummer = %d, teller = %d\nop/na regel %s\n", token_nr, token_counter, line);*/
assert(token_nr < token_counter);
 if (strcmp(token_ar[token_nr], token) == 0) {
  return 1;
 }
 else {
  return 0;
 }
}


static void check_space_at_eol() {
 if ((line[strlen(line)-1] == ' ') && (strcmp(line, "+ ") != 0)) {
  enter_error("Line ends with a space.");
  return;
 }
}


static void check_space_at_bol() {
 if (line[0] == ' ') {
  enter_error("Line starts with a space.");
  return;
 }
}

static void check_space_at_mid() {
 if (strstr(line, "  ") != 0) {
  enter_error("Line contains two (or more) consecutive spaces.");
  return;
 }
}

/******************************************************************************************/
/* split_string(in_string, ps1, ps2, start_at, splitchar)                                 */
/* splits string in_string in two parts, i.e. ps1 and ps2, starting at start_at,          */
/* with splitting character splitchar                                                     */
/* in : in_string, ps1, ps2, start_at, splitchar                                          */
/* out: none                                                                              */
/******************************************************************************************/
static void split_string(char * in_string, char ** ps1, char ** ps2, int start_at, char * splitchar) {
 lrc_string sub1;
 lrc_string sub2;
 int point_splitchar;

 point_splitchar = strcspn(in_string, splitchar);
 if (point_splitchar == (start_at - 1)) { /* het splitsteken staat aan het begin */
  strcpy(sub1, "");
 }
 else {
  sub1 = lrc_string_sub(in_string, start_at, point_splitchar);
 }
 if ((unsigned)point_splitchar == strlen(in_string)) { /* einde string */
  strcpy(sub2, "");
 }
 else {
  point_splitchar += 2;
  if ((unsigned)point_splitchar > strlen(in_string)) { /* einde string */
   strcpy(sub2, "");
  }
  else {
   sub2 = lrc_string_sub(in_string, point_splitchar, MAX_ID_NAME);
  }
 }
 *ps1 = sub1;
 *ps2 = sub2;
}


/******************************************************************************************/
/* operator_name_valid(op_name)                                                           */
/* looks up operator-name op_name in grammar                                              */
/* in : op_name                                                                           */
/* out: grammar operator number gram_op_nr                                                */
/******************************************************************************************/
static int operator_name_valid(char * op_name) {
 int gram_op_nr;
 gram_op_nr = lrc_prod_name_to_tag(op_name);
 if (gram_op_nr == -1) {
  undefined_gram_operator(op_name);
  return gram_op_nr;
 }
 return gram_op_nr;
}


/******************************************************************************************/
/* first_occurrence_in_etf(op_name)                                                       */
/* is op_name the first occurrence of op_name in etf-file? 1 = yes, 0 = no                */
/* in : op_name                                                                           */
/* out: 1 if op_name is first occurrence in etf-file, 0 if not                            */
/******************************************************************************************/
int first_occurrence_in_etf(char * op_name) {
 return (get_etf_op_nr_by_op_name(op_name) == -1);
}


/******************************************************************************************/
/* lookup_etf_op_nr_in_operators_section()                                                */
/* looks up etf-operator-number etf_op_nr (global variable) in map with operator data     */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/
static void lookup_etf_op_nr_in_operators_section() {
 if (strcmp(get_op_name_by_etf_op_nr(global_etf_op_nr), "-") == 0) { /* deze staat niet in operators section */
  etf_op_nr_not_in_operators(global_etf_op_nr);
  return;
 }
}


/******************************************************************************************/
/* lookup_string_nr_in_strings_section(key)                                               */
/* looks up string-number in map with strings data                                        */
/* in : string-number key                                                                 */
/* out: none                                                                              */
/******************************************************************************************/
static void lookup_string_nr_in_strings_section(int key) {
/* if (strcmp(get_string(key), "-1") == 0) { */
 if (!map_i_in(key, strings)) { /* zit niet in string section */
  string_not_in_string_section(key);
  return;
 }
}


/******************************************************************************************/
/* node_properties_match_with_operators_section(etf_op_nr, num_attributes, is_primitive)  */
/* checks wether num_attributes and is_primitive have appeared in operators section       */
/* in : num_attributes, is_primitive                                                      */
/* out: none                                                                              */
/******************************************************************************************/
static void node_properties_match_with_operators_section(int etf_op_nr, int num_attributes, int is_primitive) {
 if (strcmp(get_op_name_by_etf_op_nr(global_etf_op_nr), "-") != 0) { /* deze staat in operators section */
  if (num_attributes != -1) {
/*   if (num_attributes != map_ii_apply(node_properties.op_nr_num_attributes, etf_op_nr)) {
    enter_error("Parameter num_attributes in object section is not equal to num_attributes in operators section.");
    return;
   }*/
/* er vanuit gaande dat een verschil in aantal attributen niet foutief is ... ??? */

  }
  if (is_primitive != -1) {
   if (is_primitive != map_ii_apply(node_properties.op_nr_is_primitive, etf_op_nr)) {
    enter_error("Parameter is_primitive in object section is not equal to is_primitive in operators section.");
    return;
   }
  }
 }
}


static void compare_etf_with_grammar(int gram_op_nr, int num_children, int num_attributes, int is_primitive, char * category) {
 /* deze functie vergelijkt de input met de grammatica */

 /* als een onderdeel niet gecheckt hoeft te worden komt de variabele binnen met -1 of "-" */
 if (num_children != -1) {
  if (num_children != lrc_constructor_num_sons(gram_op_nr)) {
/*  printf("het is/zijn er %d", lrc_constructor_num_sons(gram_op_nr));*/
   num_children_not_conform_grammar(num_children, lrc_constructor_num_sons(gram_op_nr));
   return;
  }
 }

 if (num_attributes != -1) { /* kan niet getest worden */
 }

 if (is_primitive != -1) { /* kan niet getest worden */
 }

 if (strcmp(category, "-") != 0) {
  if (isdigit(category[0])) { /* een nummer */
   if (strcmp(get_category(atoi(category)), "0") == 0) { /* niet te vinden in categories section */
    category_not_in_categories_section(atoi(category));
    return;
   }
   if (lrc_id_to_nont(get_category(atoi(category))) == -1) {
    category_no_nonterminal(get_category(atoi(category)));
    return;
   }
  }
  else if (lrc_id_to_nont(category) == -1) {
   category_no_nonterminal(category);
   return;
  }
 }
}


static void compare_num_children_with_list_num_children(char * list_op_name, int list_num_children) {
int num_children_in_grammar;

 num_children_in_grammar = lrc_constructor_num_sons(get_gram_op_nr_by_etf_op_nr(global_etf_op_nr));
 if (num_children_in_grammar != list_num_children) {
  num_children_not_correct(list_op_name, list_num_children, num_children_in_grammar);
  return;
 }
}

 
static void convert_prim_val(char * prim_val) {
/* LET OP LET OP */
/* deze functie komt wel heelhuids door de compiler, maar is nog niet af */
/* 'k heb e.e.a. gekopieerd uit het Perl-script ETFIO.PM, en dat (deels) omgewerkt */

 /*printf("nr %d string %s\n", global_etf_op_nr, prim_val);*/
 unsigned int ix = 0;
 char letter;
 int found_slash = 0;
 int found_hex = 0;
 char character;
 char * conv_prim_val;
 char * hex_part;
 
 conv_prim_val = copy_string(prim_val);
 
/* printf(">>%s\n", prim_val);*/
 for (ix = 0; ix < strlen(prim_val); ix++) {
  letter = prim_val[ix];
  /* test op het vinden van n \   */
  if (found_slash) {
   if (found_hex) {
    if ((letter >= '0' && letter <= '9') ||
        (letter >= 'a' && letter <= 'z') ||
        (letter >= 'A' && letter <= 'Z')
       ) {
/*     printf("hex gevonden\n");*/
     if (conv_prim_val) { /* hier nog fout */
     /* =~ m/(\\[0-9a-fA-F]{2})/)*/
      character = convert_hex(hex_part);
     }
     else {
      enter_error("Impossible error situation.");
      return;
     }
     /* vervangen. nog fout */
     /* conv_prim_val =~ s/(\\[0-9a-fA-F]{2})/character/;*/
     found_slash = 0;
     found_hex = 0;
    }
    else {
     enter_error("String (primitive value) not OK: slash not followed by hex value.");
     return;
    }
   }
   else {
    if (letter == '\\') {
     /* er is al een slash, en nu nog n, dus vervangen door n \ in de converted primval */
     /* hier fout */
     /*conv_prim_val =~ s/\\\\/\\/*/
     found_slash = 0;
    }
    else {
     if ((letter >= '0' && letter <= '9') ||
         (letter >= 'a' && letter <= 'z') ||
         (letter >= 'A' && letter <= 'Z')
        ) {
      found_hex = 1;
     }
     else {
      enter_error("String (primitive value) not OK: slash not followed by hex value.");
      return;
     }
    }
   } 
  }
  else {
   if (letter == '\\') {
    found_slash = 1;
   }
  }
 }
 if (found_hex && found_slash) {
  enter_error("String (primitive value) not OK: slash followed by incomplete hex value.");
  return;
 }
 if (found_slash) {
  enter_error("String (primitive value) not OK: no slash or double slash & hex value at end of line.");
  return;
 }

/* strcpy(prim_val, conv_prim_val);*/
 prim_val = copy_string(conv_prim_val);
/* printf("Einde: |%s|.\n", conv_prim_val);*/
}


char convert_hex(char * part) {
/* $hex = substr($_[0],1,2); */
/* $int = hex($hex); */
/* helaas, nog niet af ...*/
/* return chr(hex(part[0]));*/
return 'x';
}


static void set_prim_etf_op_nr(int etf_op_nr) {
 if (get_gram_op_nr_by_etf_op_nr(etf_op_nr) == lrc_prod_name_to_tag("_Char")) {
/* if (strcmp(op_name, "_Char") == 0) {*/
  char_operator = etf_op_nr;
 }
 else if (get_gram_op_nr_by_etf_op_nr(etf_op_nr) == lrc_prod_name_to_tag("_Str")) {
/* else if (strcmp(op_name, "_Str") == 0) {*/
  str_operator = etf_op_nr;
 }
 else if (get_gram_op_nr_by_etf_op_nr(etf_op_nr) == lrc_prod_name_to_tag("_Int")) {
/* else if (strcmp(op_name, "_Int") == 0) {*/
  int_operator = etf_op_nr;
 }
 else if (get_gram_op_nr_by_etf_op_nr(etf_op_nr) == lrc_prod_name_to_tag("_Bool")) {
/* else if (strcmp(op_name, "_Bool") == 0) {*/
 bool_operator = etf_op_nr;
 }
 else if (get_gram_op_nr_by_etf_op_nr(etf_op_nr) == lrc_prod_name_to_tag("_Real")) {
/* else if (strcmp(op_name, "_Real") == 0) {*/
  real_operator = etf_op_nr;
 }
 else {
  str_operator = etf_op_nr;
/*  enter_error("Unknown operator-type for primitive value.");*/
/* ??? hier staan geen COBOL-primitieven in !!! */
  return;
 }
}


/*************** SETS & GETS ***************/
void set_version(char * version) {
 assert(version != NULL);
 final_etf.etf_version = copy_string(version); /* moet met copy_string */
}


static char * get_version() {
return "nothing";
}


static void set_tree_property(char * key, char * value) {
/*printf("key is: %s, value is: %s\n", key, value);*/
 assert(strcmp(key, FILLER_STRING) != NULL);
 assert(strcmp(value, FILLER_STRING) != NULL);
 assert(key != NULL);
 assert(value != NULL);
 final_etf.tree_properties = map_ss_set(final_etf.tree_properties, copy_string(key), copy_string(value));
}


static void get_tree_property() {
}


static void set_string(char * s) {
 assert(vrije_string_ix >= 0);
 assert(s != NULL);
 strings = map_is_set(strings, vrije_string_ix, copy_string(s));
}


static char * get_string(int key) {
 return map_is_apply(strings, key);
}


static void set_category(char * value) {
 assert(strcmp(value, FILLER_STRING) != NULL);
 assert(value != NULL);
 categories = map_is_set(categories, vrije_category_ix, copy_string(value));
}


static char * get_category(int key){
 return map_is_apply(categories, key);
}


static void set_category_properties(char * category, char * properties) {
 assert(strcmp(category, FILLER_STRING) != NULL);
 assert(strcmp(properties, FILLER_STRING) != NULL);
 assert(category != NULL);
 assert(properties != NULL);
 category_properties = map_ss_set(category_properties, copy_string(category), copy_string(properties));
}


static char * get_category_properties(char * category){
 return map_ss_apply(categories, category);
}

static void set_operator(char * op_name, int etf_op_nr, int gram_op_nr) {
 assert(strcmp(op_name, FILLER_STRING) != NULL);
 assert(strcmp(op_name, "") != NULL);
 assert(op_name != NULL);
 assert(etf_op_nr >= 0);
 assert(gram_op_nr >= 0);
 /* map vullen met etf-operator-nummer en grammar-operator-nummer */
 set_etf_op_nr_to_gram_op_nr(etf_op_nr, gram_op_nr);
 /* map vullen met operator-naam en etf-operator-nummer */
 set_op_name_to_etf_op_nr(op_name, etf_op_nr);
 if (operators_section) {
  /* map vullen met etf-operator-nummer (=volgnr binnen operators section) en operator-naam*/
  set_etf_op_nr_to_op_name(etf_op_nr, op_name);
 }
 vrije_op_ix++;
}


static void get_operator() {
}

/* set_node_properties(operator-nummer,num_children,num_attributes,is_primitive,category)
   vult alle operator-mappen
*/
static void set_node_properties(int etf_op_nr, int num_children, int num_attributes, int is_primitive, char * category) {
 assert(etf_op_nr >= 0);
 assert(num_children >= 0);
 assert(num_attributes >= 0);
 assert(is_primitive >= 0);
 assert(strcmp(category, FILLER_STRING) != NULL);
 assert(strcmp(category, "") != NULL);
 assert(category != NULL);
 
/* printf("Set operator: %d %d %d %d %s|\n", op_nr, num_children, num_attributes, is_primitive, category);*/
  set_etf_op_nr_num_children(etf_op_nr, num_children);
 if (num_attributes != -1) {
 set_etf_op_nr_num_attributes(etf_op_nr, num_attributes);
 }
 if (is_primitive != -1) {
 set_etf_op_nr_is_primitive(etf_op_nr, is_primitive);
 }
 /* bewaren van etf_op_nr van de primitieven (in primitief-soort-veld), 
    om bij het vullen van de term te weten om wat voor soort primval het gaat */
 /* printf("primi = %d\n", is_primitive);*/
 if (is_primitive) {
  set_prim_etf_op_nr(etf_op_nr);
  if (final_etf.status == 0) {return;}
 }
 if (strcmp(category,"-") != 0) {
  /*printf("hier category %s\n", category);*/
  if (isdigit(category[0])) { /* een nummer */
  set_etf_op_nr_category(etf_op_nr, get_category(atoi(category)));
  }
  else {
  set_etf_op_nr_category(etf_op_nr, category);
  }
 }
}


static void get_node_properties() {
}


static void set_etf_op_nr_to_gram_op_nr(int etf_op_nr, int gram_op_nr) {
 etf_op_nr_to_gram_op_nr = map_ii_set(etf_op_nr_to_gram_op_nr, etf_op_nr, gram_op_nr);
}


static int get_gram_op_nr_by_etf_op_nr(int etf_op_nr) {
 return map_ii_apply(etf_op_nr_to_gram_op_nr, etf_op_nr);
}


static void set_op_name_to_etf_op_nr(char * op_name, int etf_op_nr) {
 assert(strcmp(op_name, FILLER_STRING) != NULL);
 assert(strcmp(op_name, "") != NULL);
 assert(op_name != NULL);
 assert(etf_op_nr >= 0);
 
 op_name_to_etf_op_nr = map_si_set(op_name_to_etf_op_nr, copy_string(op_name), etf_op_nr);
/*printf("set \n");
lrc_map_print(op_name_to_etf_op_nr,lrc_term_print_long,lrc_term_print_long);*/
}


static int get_etf_op_nr_by_op_name(char * op_name) {
/*printf("get \n");
lrc_map_print(op_name_to_etf_op_nr,lrc_term_print_long,lrc_term_print_long);*/
 return map_si_apply(op_name_to_etf_op_nr, op_name);
}


static void set_etf_op_nr_to_op_name(int etf_op_nr, char * op_name) {
 assert(etf_op_nr >= 0);
 assert(strcmp(op_name, FILLER_STRING) != NULL);
 assert(strcmp(op_name, "") != NULL);
 assert(op_name != NULL);
 etf_op_nr_to_op_name = map_is_set(etf_op_nr_to_op_name, etf_op_nr, copy_string(op_name));
}


static char * get_op_name_by_etf_op_nr(int etf_op_nr) {
 return map_is_apply(etf_op_nr_to_op_name, etf_op_nr);
}


static void set_etf_op_nr_num_children(int etf_op_nr, int num_children) {
 assert(etf_op_nr >= 0);
 assert(num_children >= 0);
 node_properties.op_nr_num_children = map_ii_set(node_properties.op_nr_num_children, etf_op_nr, num_children);
}


static int get_etf_op_nr_num_children(int etf_op_nr) {
 return map_ii_apply(node_properties.op_nr_num_children, etf_op_nr);
}


static void set_etf_op_nr_num_attributes(int etf_op_nr, int num_attributes) {
 assert(etf_op_nr >= 0);
 assert(num_attributes >= 0);
 node_properties.op_nr_num_attributes = map_ii_set(node_properties.op_nr_num_attributes, etf_op_nr, num_attributes);
}


static int get_etf_op_nr_num_attributes(int etf_op_nr) {
/*printf("get \n");
lrc_map_print(node_properties.op_nr_num_attributes,lrc_term_print_long,lrc_term_print_long);*/
 return map_ii_apply(node_properties.op_nr_num_attributes, etf_op_nr);
}


static void set_etf_op_nr_is_primitive(int etf_op_nr, int is_primitive) {
 assert(etf_op_nr >= 0);
 assert(is_primitive >= 0);
 node_properties.op_nr_is_primitive = map_ii_set(node_properties.op_nr_is_primitive, etf_op_nr, is_primitive);
}


static int get_etf_op_nr_is_primitive(int etf_op_nr) {
 return map_ii_apply(node_properties.op_nr_is_primitive, etf_op_nr);
}


static void set_etf_op_nr_category(int etf_op_nr, char * category) {
 assert(etf_op_nr >= 0);
 assert(strcmp(category, FILLER_STRING) != NULL);
 assert(strcmp(category, "") != NULL);
 assert(category != NULL);
 node_properties.op_nr_category = map_is_set(node_properties.op_nr_category, etf_op_nr, copy_string(category));
}


static void get_etf_op_nr_category() {
}


static void enter_string() {
 /* aannemen dat string-regels goed zijn.*/
 if (line[0] == '$' && line[1] == '$') {
  set_string(line + 1);
 }
 else {
  set_string(line);
 }
 vrije_string_ix++;
}


static void enter_category(char * category, char * properties) {
 set_category(category);
 vrije_category_ix++;
 set_category_properties(category, properties);
}



/*************** ERROR MESSAGES ***************/
static void enter_error(char * error_message) {
 final_etf.status = 0;
 sprintf(final_etf.message, error_message);
}


static void undefined_gram_operator(char * op_name) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s", "The operator-name '", op_name, "' is not a production in the grammar.");
 enter_error(message);
}


static void category_no_nonterminal(char * category) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s", "Category '", category, "' is not a nonterminal in the grammar.");
 enter_error(message);
}


static void category_no_root(char * category, char * category_properties) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s%s%s", "Category-properties '", category_properties, "' indicate root with R, but category '", category , "' is not a root in the grammar.");
 enter_error(message);
}


static void name_no_identifier(char * this, char * op_name) { /* deze is, i.v.m. controle op grammatica, overbodig /??? */
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s%s%s", "This ", this, "'", op_name, "' is not an identifier.");
 enter_error(message);
}


static void op_name_not_in_operators(char * op_name) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s", "The operator-name '", op_name, "' is not listed in the operators section.");
 enter_error(message);
}


static void etf_op_nr_not_in_operators(int etf_op_nr) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%d%s", "There is no operator-name associated with etf-operator-number ", etf_op_nr, ".");
 enter_error(message);
}


static void num_children_not_correct(char * op_name, int right_num, int wrong_num) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%s%s%d%s%d%s", "Number of children for list-operator ", op_name, " should be ", right_num, " instead of ", wrong_num, ".");
 enter_error(message);
}


static void num_children_not_conform_grammar(int etf_num_children, int gram_num_children) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%d%s%d%s", "Number of children (", etf_num_children, ") not conform grammar (", gram_num_children, ").");
 enter_error(message);
}


static void category_not_in_categories_section(int cat_nr) {
char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%d%s", "Lookup in Categories section: there is no category associated with number ", cat_nr, ".");
 enter_error(message);
}
 

static void string_not_in_string_section(int str_nr) {
 char * message = (lrc_string)ALLOC_STR(MAXLINE);
 sprintf(message, "%s%d%s", "Lookup in Strings section: there is no string associated with number ", str_nr, ".");
 enter_error(message);
}


/******************************************************************************************/
/*                                                                                        */
/*                                                                                        */
/* in : none                                                                              */
/* out: none                                                                              */
/******************************************************************************************/

