#include <stdio.h>
/* #include <unistd.h> */
#include <string.h>
#include <ctype.h>
#include "lrcmemory.h"
#include "ag_table.h"

static char * line;

#define MAX_TOKENS 12

static int token_length[MAX_TOKENS];
static int token_start[MAX_TOKENS];
static int num_tokens;

/* tokenize_line takes a line and determines the
   start-index and length of all tokens in that line.
   A token is a sequence of [A-Za-z_] or a sequence of [0-9].
   Furthermore the number of tokens in the line is set.

   The line is assumed to contain a final '\n'.
 */

static void print_line() {
  printf("%s",line); fflush(stdout);
}

static int is_token_char(char c) { 
  return (isalpha(c) || isdigit(c) || c=='_');
}

static int is_end_char(char c) {
  return c=='\0' || c=='\n';
}

static int is_comment_line(char * line) {
  return line[0]=='#' || line[0]=='/';
}

static void tokenize_line(char * line) {
  int len;
  int cur_token;
  int index;

  num_tokens=0;

  /** printf("Processing: "); print_line(line); **/
  len=strlen(line);
  if (len==0 || len==1) return;
  if (is_comment_line(line)) return;
  index=0;
  cur_token=0;
  while(index<len-1) {
    while (!is_token_char(line[index]) && !is_end_char(line[index])) index++;
    if (is_token_char(line[index])) {
      token_start[cur_token]=index;
      while(is_token_char(line[index])) {
        index++;
      }
      token_length[cur_token]=index-token_start[cur_token];
      cur_token++;
    }
  }
  num_tokens=cur_token;
}

static char index_token(int tok_num, int ix) {
   if (tok_num>=num_tokens) printf("index_token:Illegal token number %d\n", tok_num);
   if (ix>=token_length[tok_num]) printf("index_token:Illegal index %d\n", ix);
   return line[token_start[tok_num]+ix];
}

static char * token_to_string(int tok_number) {
  char * res;
  int i;
  if (tok_number>=num_tokens) {
     printf("token_to_string:wrong token number:%d\n",tok_number); fflush(stdout);
     print_line();
     return "";
  }
  res=(char *)ALLOC_STR(token_length[tok_number]+1);
  for(i=0;i<token_length[tok_number];i++) res[i]=index_token(tok_number,i);
  res[token_length[tok_number]]='\0';
  return res;
}

static int token_to_int(int tok_number) {
  int res=0;
  int i;
  if (tok_number>=num_tokens) {
     printf("token_to_int:wrong token number:%d\n",tok_number); fflush(stdout);
     print_line();
     return 0;
  }
  if (!isdigit(index_token(tok_number,0))) {
     printf("token_to_int:token does not start with a digit:%d(%c)\n",
                                  tok_number,index_token(tok_number,0));
     fflush(stdout);
     print_line();
     return 0;
  }
  for(i=0;i<token_length[tok_number];i++) res=10*res+index_token(tok_number,i)-'0';
  return res;
}

static int token_equal_string(int token_number, char * s) {
  int difference;
  int i;

  if (token_length[token_number]!=strlen(s)) return 0;
  difference=0;
  i=0;
  for(i=0;i<token_length[token_number] && !difference;i++) {
    difference=index_token(token_number,i)!=s[i];
  }
  return !difference;
}

static void print_tokens() {
  int tok_num, tix;
  for(tok_num=0;tok_num<num_tokens;tok_num++) {
     printf("%d:", tok_num); fflush(stdout);
     for(tix=0; tix<token_length[tok_num]; tix++) {
        printf("%c", index_token(tok_num,tix));
     }
     printf("\n"); fflush(stdout);
  }
}

static int current_ident_ix=0;
static void proces_start_idents() {
  int num_idents;
  num_idents=token_to_int(1);
  /*** printf("proces_start_idents %d\n",num_idents); ***/
  allocate_attr_identifier_data(num_idents);
  current_ident_ix=0;
}

static void proces_lrc_ident() {
  /* printf("proces_lrc_ident %d %s\n", token_to_int(1), token_to_string(2)); */
  define_attr_ident(current_ident_ix,token_to_int(1),token_to_string(2));
  current_ident_ix++;
}

static void proces_end_idents() {
 /*** printf("proces_end_idents\n"); ***/
 /* print_attr_idents(); */
}

/* ============  Abstract Syntax ======== */

static int current_nt=0;

static void  proces_nont() {
  /*printf("proces_lrc_nont %d %s\n", token_to_string(1)); fflush(stdout);*/
  enter_nonterminal(current_nt,token_to_string(1));
  current_nt++;
}

static void  proces_prod() {
  /*** printf("Processing production %s\n", token_to_string(1)); fflush(stdout); ***/
  enter_production(token_to_string(1),token_to_int(2),token_to_int(3),token_to_int(4));
}

static void  proces_ntocc() {
  enter_ntocc(token_to_int(1),token_to_int(2),token_to_int(3),token_to_int(4),token_to_int(5));
}

static void  proces_start_nonterminals() {
  /** printf("proces_start_nonterminals %d\n",token_to_int(1)); fflush(stdout); **/
  allocate_nonterminal_data(token_to_int(1));
}

static void  proces_end_nonterminals() {
  /** printf("proces_end_nonterminals\n"); fflush(stdout); **/
}

static void  proces_start_productions() {
  /** printf("proces_start_productions\n"); fflush(stdout); **/
  allocate_production_data(token_to_int(1));
}

static void  proces_end_productions() {
  /** printf("proces_end_productions\n"); fflush(stdout); **/
}

static void  proces_start_ntoccs() {
  /** printf("proces_start_ntoccs\n"); fflush(stdout); **/
  allocate_ntocc_data(token_to_int(1));
}

static void  proces_end_ntoccs() {
  /** printf("proces_end_ntoccs\n"); fflush(stdout); **/
}

static void proces_ids_tokens() {
  if (token_equal_string(0,"LRC_IDENTIFIERS_START")) {
     proces_start_idents();
  }
  else if (token_equal_string(0,"ID")) {
     proces_lrc_ident();
  }
  else if (token_equal_string(0,"LRC_IDENTIFIERS_END")) {
     proces_end_idents();
  }
}

static void proces_das_tokens() {
       if (token_equal_string(0,"N"))                      { proces_nont();   }
  else if (token_equal_string(0,"P"))                      { proces_prod();   }
  else if (token_equal_string(0,"NO"))                     { proces_ntocc();   }
  else if (token_equal_string(0,"LRC_START_NONTERMINALS")) { proces_start_nonterminals();   }
  else if (token_equal_string(0,"LRC_END_NONTERMINALS"))   { proces_end_nonterminals();   }
  else if (token_equal_string(0,"LRC_START_PRODUCTIONS"))  { proces_start_productions();   }
  else if (token_equal_string(0,"LRC_END_PRODUCTIONS"))    { proces_end_productions();   }
  else if (token_equal_string(0,"LRC_START_NTOCCS"))       { proces_start_ntoccs();   }
  else if (token_equal_string(0,"LRC_END_NTOCCS"))         { proces_end_ntoccs();   }
}

/* ===== Semantics =========== */
/*        AA  global_aocc_number ntocc-number attribute-number */
/*  token 0      1                2            3               */
static void proces_aocc_a(){
  enter_aocc_a(token_to_int(1),token_to_int(2),token_to_int(3));
}

/*        AN  global_aocc_number ntocc-number */
/* token  0    1                  2           */
static void proces_aocc_n(){
  enter_aocc_n(token_to_int(1),token_to_int(2));
}

/*        AL  global_aocc_number prod-number attribute-number */
/*  token 0     1                   2             3           */
static void proces_aocc_l(){
  enter_aocc_l(token_to_int(1),token_to_int(2),token_to_int(3));
}

/*       L global_attr_num attr-id  production-number */
/* token 0   1           2              3             */
static void proces_prod_attr(){
  enter_attribute(token_to_int(2),token_to_int(1),token_to_int(3),LOCAL);
}

/*       I   global_attr_num attr-id  nonterminal-number */
/*       S   global_attr_num attr-id  nonterminal-number */
/* token 0      1              2           3 */
static void proces_nont_attr(){
  attr_kind_tp kind;
  if (token_equal_string(0,"I")) kind=INH; else kind=SYN;
  enter_attribute(token_to_int(2),token_to_int(1),token_to_int(3),kind);
}

/*       E equation_number destination-aocc  */
/* token 0     1                2 */
static void proces_equation() {
  enter_eq(token_to_int(1),token_to_int(2));
}

static void proces_start_attributes(){
  /** printf("proces_start_attributes %d\n",token_to_int(1)); fflush(stdout); **/
  allocate_attribute_data(token_to_int(1));
}

static void proces_end_attributes(){
  /** printf("proces_end_attributes\n"); fflush(stdout); **/
}

static void proces_start_aoccs(){
  /** printf("proces_start_aoccs %d\n",token_to_int(1)); fflush(stdout); **/
  allocate_aocc_data(token_to_int(1));
}

static void proces_end_aoccs(){
  /** printf("proces_end_aoccs\n"); fflush(stdout); **/
}

static void proces_start_eqs(){
  /** printf("proces_start_eqs %d\n",token_to_int(1)); fflush(stdout); **/
  allocate_eq_data(token_to_int(1));
}

static void proces_end_eqs(){
  /** printf("proces_end_eqs\n"); fflush(stdout); **/
}


static void proces_dsm_tokens() {
       if (token_equal_string(0,"AA"))             { proces_aocc_a();   }
  else if (token_equal_string(0,"AL"))             { proces_aocc_l();   }
  else if (token_equal_string(0,"AN"))             { proces_aocc_n();   }
  else if (token_equal_string(0,"L"))              { proces_prod_attr();   }
  else if (token_equal_string(0,"I"))              { proces_nont_attr();   }
  else if (token_equal_string(0,"S"))              { proces_nont_attr();   }
  else if (token_equal_string(0,"E"))              { proces_equation();    }
  else if (token_equal_string(0,"LRC_START_ATTRIBUTES")) { proces_start_attributes();   }
  else if (token_equal_string(0,"LRC_END_ATTRIBUTES"))   { proces_end_attributes();   }
  else if (token_equal_string(0,"LRC_START_AOCCS"))      { proces_start_aoccs();   }
  else if (token_equal_string(0,"LRC_END_AOCCS"))        { proces_end_aoccs();   }
  else if (token_equal_string(0,"LRC_START_EQS"))        { proces_start_eqs();   }
  else if (token_equal_string(0,"LRC_END_EQS"))          { proces_end_eqs();   }
}

static void proces_file(char * fname, void token_processor() ) {
  FILE *f;

  f=fopen(fname,"r");
  if (!f) {
    printf("Cannot open %s\n",fname); return;
  }

  line=(char *)lrc_malloc(401);
  while(fgets(line,400,f)) {
     /* printf("%s", line); fflush(stdout); */
     tokenize_line(line);
     /* print_tokens(); */
     if (num_tokens>0) token_processor();
  }
}

/* === Locating data files === */
static char * dir_file_concat(char * s0, char * s1) {
  char * res;
  res=(char *) ALLOC_STR(strlen(s0)+1+strlen(s1)+1);
  res[0]=0;
  strcat(res,s0); strcat(res,"/"); strcat(res,s1);
  return res;
}

/* access returns 0(zero) on success, -1 otherwise */
static int readable(char * fname) {
  /* return (access(fname,R_OK)==0); */ /* MFK: access does not compiler under Borland; */
  FILE *f;
  f=fopen(fname,"r");
  if (!f) {
    return 0;
  }
  fclose(f);
  return 1; 
}

static int directory_contains_file(char * dir_name, char * fname) {
  return readable(dir_file_concat(dir_name,fname));
}

static int directory_contains_files(char * dir_name, char ** files) {
  int ok, ix;
  ok=1;
  ix=0;
  while (ok && files[ix]!=0) {
    ok=directory_contains_file(dir_name,files[ix]);
    if (ok) ix++;
  }
  return ok && files[ix]==0;
}

/* Locate a directory in path that contains all files from files.
 *  path and files are zero-terminated lists of names.
 *
 * Returns the directory that contains the files, or 0 if no directory in
 *  the path contains the files.
 */
char * lrc_locate_dir(char ** path, char **files) {
  int ix, found;
  ix=0; found=0;
  while (path[ix] && !found) {
    found=directory_contains_files(path[ix],files);
    if (!found) ix++;
  }
  return path[ix];
}

/* The data of the ag is read from the following files: */
/* static char * data_files[] = { "code.ids", "code.das", "code.dsm", 0 }; */
static char * data_files[] = { "code.ids", "code.dsm", 0 };

int lrc_read_ag_tables(char ** search_path) {
  char * code_dir;

  code_dir=lrc_locate_dir(search_path,data_files);
  if (!code_dir) return 0;
  proces_file(dir_file_concat(code_dir,"code.ids"),proces_ids_tokens);
  proces_file(dir_file_concat(code_dir,"code.das"),proces_das_tokens);
  proces_file(dir_file_concat(code_dir,"code.dsm"),proces_dsm_tokens);
  return 1;
}
